# MatrixZulipBridge - an appservice puppeting bridge for Matrix - Zulip
#
# Copyright (C) 2024 Emma Meijere <emgh@em.id.lv>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Originally licensed under the MIT (Expat) license:
# <https://github.com/hifi/heisenbridge/blob/2532905f13835762870de55ba8a404fad6d62d81/LICENSE>.
#
# [This file includes modifications made by Emma Meijere]
#
#
import asyncio
import logging
import re
from typing import TYPE_CHECKING, Optional
from urllib.parse import urlparse

from bs4 import BeautifulSoup
from markdownify import markdownify
from mautrix.api import Method, SynapseAdminPath
from mautrix.errors import MatrixStandardRequestError
from mautrix.types.event.state import (
    JoinRestriction,
    JoinRestrictionType,
    JoinRule,
    JoinRulesStateEventContent,
)
from mautrix.types.event.type import EventType

from matrixzulipbridge.command_parse import (
    CommandManager,
    CommandParser,
    CommandParserError,
)
from matrixzulipbridge.room import Room

if TYPE_CHECKING:
    from matrixzulipbridge.organization_room import OrganizationRoom


def connected(f):
    def wrapper(*args, **kwargs):
        self = args[0]

        if (
            not self.organization
            or not self.organization.conn
            or not self.organization.conn.connected
        ):
            self.send_notice("Need to be connected to use this command.")
            return asyncio.sleep(0)

        return f(*args, **kwargs)

    return wrapper


class PrivateRoom(Room):
    name: str
    organization: Optional["OrganizationRoom"]
    organization_id: str
    organization_name: Optional[str]
    media: list[list[str]]
    max_backfill_amount: int

    force_forward = False

    commands: CommandManager

    def init(self) -> None:
        self.name = None
        self.organization = None
        self.organization_id = None
        self.organization_name = None  # deprecated
        self.media = []
        self.lazy_members = {}  # allow lazy joining your own ghost for echo
        self.max_backfill_amount = None

        self.commands = CommandManager()

        if type(self) == PrivateRoom:
            cmd = CommandParser(prog="WHOIS", description="WHOIS the other user")
            self.commands.register(cmd, self.cmd_whois)

        cmd = CommandParser(
            prog="BACKFILL",
            description="set the maximum amount of backfilled messages (0 to disable backfilling)",
        )
        cmd.add_argument("amount", nargs="?", help="new amount")
        self.commands.register(cmd, self.cmd_backfill)

        self.mx_register("m.room.message", self.on_mx_message)
        self.mx_register("m.room.redaction", self.on_mx_redaction)

    def from_config(self, config: dict) -> None:
        super().from_config(config)

        if "name" not in config:
            raise Exception("No name key in config for ChatRoom")

        self.name = config["name"]

        if "organization_id" in config:
            self.organization_id = config["organization_id"]

        if "media" in config:
            self.media = config["media"]

        # only used for migration
        if "organization" in config:
            self.organization_name = config["organization"]

        if self.organization_name is None and self.organization_id is None:
            raise Exception(
                "No organization or organization_id key in config for PrivateRoom"
            )

        if "max_backfill_amount" in config:
            self.max_backfill_amount = config["max_backfill_amount"]

    def to_config(self) -> dict:
        return {
            **(super().to_config()),
            "name": self.name,
            "organization": self.organization_name,
            "organization_id": self.organization_id,
            "media": self.media[:5],
            "max_backfill_amount": self.max_backfill_amount,
        }

    @staticmethod
    async def create(organization: "OrganizationRoom", name: str) -> "PrivateRoom":
        logging.debug(
            f"PrivateRoom.create(organization='{organization.name}', name='{name}')"
        )
        raise NotImplementedError("Direct messaging")

        # asyncio.ensure_future(room.create_mx(name))
        # return room

    async def create_mx(self, displayname) -> None:
        if self.id is None:
            mx_user_id = await self.organization.serv.ensure_zulip_user_id(
                self.organization.name, displayname, update_cache=False
            )
            self.id = await self.organization.serv.create_room(
                f"{displayname} ({self.organization.name})",
                "Private chat with {displayname} on {self.organization.name}",
                [self.organization.user_id, mx_user_id],
            )
            self.serv.register_room(self)
            await self.az.intent.user(mx_user_id).ensure_joined(self.id)
            await self.save()
            # start event queue now that we have an id
            self._queue.start()

            # attach to organization space
            if self.organization.space:
                await self.organization.space.attach(self.id)

    def is_valid(self) -> bool:
        if self.organization_id is None and self.organization_name is None:
            return False

        if self.name is None:
            return False

        if self.user_id is None:
            return False

        if not self.in_room(self.user_id):
            return False

        return True

    def cleanup(self) -> None:
        logging.debug(f"Cleaning up organization connected room {self.id}.")

        # cleanup us from organization space if we have it
        if self.organization and self.organization.space:
            asyncio.ensure_future(self.organization.space.detach(self.id))

        # cleanup us from organization rooms
        if self.organization and self.name in self.organization.rooms:
            logging.debug(
                f"... and we are attached to organization {self.organization.id}, detaching."
            )
            del self.organization.rooms[self.name]
        super().cleanup()

    def send_notice(
        self,
        text: str,
        user_id: Optional[str] = None,
        formatted=None,
        fallback_html: Optional[str] = None,
        forward=False,
    ):
        if (
            self.force_forward or forward or self.organization.forward
        ) and user_id is None:
            self.organization.send_notice(
                text=f"{self.name}: {text}",
                formatted=formatted,
                fallback_html=fallback_html,
            )
        else:
            super().send_notice(
                text=text,
                user_id=user_id,
                formatted=formatted,
                fallback_html=fallback_html,
            )

    def send_notice_html(
        self, text: str, user_id: Optional[str] = None, forward=False
    ) -> None:
        if (
            self.force_forward or forward or self.organization.forward
        ) and user_id is None:
            self.organization.send_notice_html(text=f"{self.name}: {text}")
        else:
            super().send_notice_html(text=text, user_id=user_id)

    def pills(self):
        # if pills are disabled, don't generate any
        if self.organization.pills_length < 1:
            return None

        ret = {}
        ignore = list(map(lambda x: x.lower(), self.organization.pills_ignore))

        # push our own name first
        lnick = self.organization.conn.real_nickname.lower()
        if (
            self.user_id in self.displaynames
            and len(lnick) >= self.organization.pills_length
            and lnick not in ignore
        ):
            ret[lnick] = (self.user_id, self.displaynames[self.user_id])

        # assuming displayname of a puppet matches nick
        for member in self.members:
            if not self.serv.is_puppet(member):
                continue

            if member in self.displaynames:
                nick = self.displaynames[member]
                lnick = nick.lower()
                if len(nick) >= self.organization.pills_length and lnick not in ignore:
                    ret[lnick] = (member, nick)

        return ret

    async def _process_event_content(self, event, prefix="", _reply_to=None):
        content = event.content

        if content.formatted_body:
            message = content.formatted_body

            if "m.mentions" in content:
                mentioned_users: list[str] = event.content["m.mentions"].get(
                    "user_ids", []
                )
                soup = BeautifulSoup(content.formatted_body, features="html.parser")
                for mxid in mentioned_users:
                    # Translate puppet mentions as native Zulip mentions
                    if not self.serv.is_puppet(mxid):
                        continue

                    user_id = self.organization.get_zulip_user_id_from_mxid(mxid)
                    zulip_user = self.organization.get_zulip_user(user_id)

                    # Replace matrix.to links with Zulip mentions
                    mentions = soup.find_all(
                        "a",
                        href=f"https://matrix.to/#/{mxid}",
                    )

                    zulip_mention = soup.new_tag("span")
                    zulip_mention.string = " @"
                    zulip_mention_content = soup.new_tag("strong")
                    zulip_mention_content.string = (
                        f"{zulip_user['full_name']}|{user_id}"
                    )
                    zulip_mention.append(zulip_mention_content)

                    for mention in mentions:
                        mention.replace_with(zulip_mention)
                message = soup.encode(formatter="html5")

            message = markdownify(message)
        elif content.body:
            message = content.body
        else:
            logging.warning("_process_event_content called with no usable body")
            return
        message = prefix + message
        return message

    async def _send_message(self, event, prefix=""):
        raise NotImplementedError("Private messages to Zulip")

    async def on_mx_message(self, event) -> None:
        if event.sender != self.user_id:
            return

        if (
            self.organization is None
            or self.organization.zulip is None
            or not self.organization.zulip.has_connected
        ):
            self.send_notice("Not connected to organization.")
            return

        if str(event.content.msgtype) == "m.emote":
            await self._send_message(event)
        elif str(event.content.msgtype) in ["m.image", "m.file", "m.audio", "m.video"]:
            self.organization.conn.privmsg(
                self.name, self.serv.mxc_to_url(event.content.url, event.content.body)
            )
            self.react(event.event_id, "\U0001F517")  # link
            self.media.append([event.event_id, event.content.url])
            await self.save()
        elif str(event.content.msgtype) == "m.text":
            # allow commanding the appservice in rooms
            match = re.match(r"^\s*@?([^:,\s]+)[\s:,]*(.+)$", event.content.body)
            if (
                match
                and match.group(1).lower() == self.serv.registration["sender_localpart"]
            ):
                try:
                    await self.commands.trigger(match.group(2))
                except CommandParserError as e:
                    self.send_notice(str(e))
                return

            await self._send_message(event)

        await self.az.intent.send_receipt(event.room_id, event.event_id)

    async def on_mx_redaction(self, event) -> None:
        for media in self.media:
            if media[0] == event.redacts:
                url = urlparse(media[1])
                if self.serv.synapse_admin:
                    try:
                        await self.az.intent.api.request(
                            Method.POST,
                            SynapseAdminPath.v1.media.quarantine[url.netloc][
                                url.path[1:]
                            ],
                        )

                        self.organization.send_notice(
                            f"Associated media {media[1]} for redacted event {event.redacts} "
                            + f"in room {self.name} was quarantined."
                        )
                    except Exception:
                        self.organization.send_notice(
                            f"Failed to quarantine media! Associated media {media[1]} "
                            + f"for redacted event {event.redacts} in room {self.name} is left available."
                        )
                else:
                    self.organization.send_notice(
                        f"No permission to quarantine media! Associated media {media[1]} "
                        + f"for redacted event {event.redacts} in room {self.name} is left available."
                    )
                return

    @connected
    async def cmd_whois(self, _args) -> None:
        self.organization.conn.whois(f"{self.name} {self.name}")

    async def cmd_backfill(self, args) -> None:
        if args.amount:
            self.max_backfill_amount = int(args.amount)
            await self.save()
        self.send_notice(
            f"Maximum backfill amount is set to: {self.max_backfill_amount}"
        )

    async def _attach_space_internal(self) -> None:
        await self.az.intent.send_state_event(
            self.id,
            EventType.ROOM_JOIN_RULES,  # Why does this happend? pylint: disable=no-member
            content=JoinRulesStateEventContent(
                join_rule=JoinRule.RESTRICTED,
                allow=[
                    JoinRestriction(
                        type=JoinRestrictionType.ROOM_MEMBERSHIP,
                        room_id=self.organization.space.id,
                    ),
                ],
            ),
        )

    async def _attach_space(self) -> None:
        logging.debug(
            f"Attaching room {self.id} to organization space {self.organization.space.id}."
        )
        try:
            room_create = await self.az.intent.get_state_event(
                self.id, EventType.ROOM_CREATE
            )  # pylint: disable=no-member
            if room_create.room_version in [str(v) for v in range(1, 9)]:
                self.send_notice(
                    "Only rooms of version 9 or greater can be attached to a space."
                )
                self.send_notice(
                    "Leave and re-create the room to ensure the correct version."
                )
                return

            await self._attach_space_internal()
            self.send_notice("Attached to space.")
        except MatrixStandardRequestError as e:
            logging.debug("Setting join_rules for space failed.", exc_info=True)
            self.send_notice(f"Failed attaching space: {e.message}")
            self.send_notice("Make sure the room is at least version 9.")
        except Exception:
            logging.exception(
                f"Failed to attach {self.id} to space {self.organization.space.id}."
            )

    async def cmd_upgrade(self, args) -> None:
        if not args.undo:
            await self._attach_space()



# MatrixZulipBridge - an appservice puppeting bridge for Matrix - Zulip
#
# Copyright (C) 2024 Emma Meijere <emgh@em.id.lv>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Originally licensed under the MIT (Expat) license:
# <https://github.com/hifi/heisenbridge/blob/2532905f13835762870de55ba8a404fad6d62d81/LICENSE>.
#
# [This file includes modifications made by Emma Meijere]
#
#
import asyncio
import logging


class EventQueue:
    def __init__(self, callback):
        self._callback = callback
        self._events = []
        self._loop = asyncio.get_running_loop()
        self._timer = None
        self._start = 0
        self._chain = asyncio.Queue()
        self._task = None
        self._timeout = 3600

    def start(self):
        if self._task is None:
            self._task = asyncio.ensure_future(self._run())

    def stop(self):
        if self._task:
            self._task.cancel()
            self._task = None

    async def _run(self):
        while True:
            try:
                task = await self._chain.get()
            except asyncio.CancelledError:
                logging.debug("EventQueue was cancelled.")
                return

            try:
                await asyncio.create_task(task)
            except asyncio.CancelledError:
                logging.debug("EventQueue task was cancelled.")
                return
            except asyncio.TimeoutError:
                logging.warning("EventQueue task timed out.")
            finally:
                self._chain.task_done()

    def _flush(self):
        events = self._events

        self._timer = None
        self._events = []

        self._chain.put_nowait(self._callback(events))

    def enqueue(self, event):
        now = self._loop.time()

        # always cancel timer when we enqueue
        if self._timer:
            self._timer.cancel()

        # stamp start time when we queue first event, always append event
        if len(self._events) == 0:
            self._start = now

        self._events.append(event)

        # if we have bumped ourself for half a second, flush now
        if now >= self._start + 0.5:
            self._flush()
        else:
            self._timer = self._loop.call_later(0.1, self._flush)



# MatrixZulipBridge - an appservice puppeting bridge for Matrix - Zulip
#
# Copyright (C) 2024 Emma Meijere <emgh@em.id.lv>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Originally licensed under the MIT (Expat) license:
# <https://github.com/hifi/heisenbridge/blob/2532905f13835762870de55ba8a404fad6d62d81/LICENSE>.
#
# [This file includes modifications made by Emma Meijere]
#
#
import argparse
import asyncio
import datetime
import functools
import html
import json
import logging
import re
from argparse import Namespace
from collections import defaultdict
from typing import TYPE_CHECKING, Any

import zulip
from mautrix.util.bridge_state import BridgeStateEvent

from matrixzulipbridge import __version__
from matrixzulipbridge.command_parse import (
    CommandManager,
    CommandParser,
    CommandParserError,
)
from matrixzulipbridge.private_room import PrivateRoom
from matrixzulipbridge.room import Room
from matrixzulipbridge.space_room import SpaceRoom
from matrixzulipbridge.stream_room import StreamRoom
from matrixzulipbridge.zulip import ZulipEventHandler

if TYPE_CHECKING:
    from matrixzulipbridge.appservice import AppService


def connected(f):
    def wrapper(*args, **kwargs):
        self = args[0]

        if not self.zulip or not self.zulip.has_connected:
            self.send_notice("Need to be connected to use this command.")
            return asyncio.sleep(0)

        return f(*args, **kwargs)

    return wrapper


class OrganizationRoom(Room):
    # configuration stuff
    name: str
    connected: bool
    fullname: str

    api_key: str
    email: str
    site: str
    zulip: "zulip.Client"
    zulip_users: dict[int, dict]

    # state
    commands: CommandManager
    rooms: dict[str, Room]
    connecting: bool
    real_host: str
    real_user: str
    pending_kickbans: dict[str, list[tuple[str, str]]]
    backoff: int
    backoff_task: Any
    next_server: int
    connected_at: int
    space: SpaceRoom
    post_init_done: bool
    caps_supported: list
    caps_enabled: list
    caps_task: Any
    disconnect: bool

    organization: Any

    profile: dict
    messages: dict[str, str]
    permissions: dict[str, str]
    zulip_handler: "ZulipEventHandler"
    max_backfill_amount: int

    def init(self):
        self.name = None
        self.connected = False
        self.fullname = None
        self.ircname = None
        self.password = None
        self.sasl_mechanism = None
        self.sasl_username = None
        self.sasl_password = None
        self.autocmd = None
        self.pills_length = 2
        self.pills_ignore = []
        self.autoquery = True
        self.allow_ctcp = False
        self.tls_cert = None
        self.rejoin_invite = True
        self.rejoin_kick = False
        self.caps = ["message-tags", "chghost", "znc.in/self-message"]
        self.forward = False
        self.backoff = 0
        self.backoff_task = None
        self.next_server = 0
        self.connected_at = 0

        self.api_key = None
        self.email = None
        self.site = None
        self.zulip_users = {}

        self.commands = CommandManager()
        self.zulip = None
        self.rooms = {}
        self.connlock = asyncio.Lock()
        self.disconnect = True
        self.real_host = "?" * 63  # worst case default
        self.real_user = "?" * 8  # worst case default
        self.keys = {}  # temp dict of join channel keys
        self.keepnick_task = None  # async task
        self.whois_data = defaultdict(dict)  # buffer for keeping partial whois replies
        self.pending_kickbans = defaultdict(list)
        self.space = None
        self.post_init_done = False
        self.caps_supported = []
        self.caps_enabled = []
        self.caps_task = None

        self.organization = self

        self.profile = None
        self.messages = {}
        self.permissions = {}
        self.zulip_handler = None
        self.max_backfill_amount = 100

        cmd = CommandParser(
            prog="FULLNAME",
            description="set/change full name",
            epilog=(
                "You can always see your current full name on the organization without arguments.\n"
            ),
        )
        cmd.add_argument("fullname", nargs="?", help="new full name")
        self.commands.register(cmd, self.cmd_fullname)

        cmd = CommandParser(
            prog="SITE",
            description="set Zulip site",
            epilog=(),
        )
        cmd.add_argument("site", nargs="?", help="new site")
        self.commands.register(cmd, self.cmd_site)

        cmd = CommandParser(
            prog="EMAIL",
            description="set Zulip bot email",
            epilog=(),
        )
        cmd.add_argument("email", nargs="?", help="new bot email")
        self.commands.register(cmd, self.cmd_email)

        cmd = CommandParser(
            prog="APIKEY",
            description="set Zulip bot api key",
            epilog=(),
        )
        cmd.add_argument("api_key", nargs="?", help="new API key")
        self.commands.register(cmd, self.cmd_apikey)

        cmd = CommandParser(
            prog="CONNECT",
            description="connect to organization",
            epilog=(
                "When this command is invoked the connection to this organization will be persisted across disconnects and"
                " bridge restart.\n"
                "Only if the server KILLs your connection it will stay disconnected until CONNECT is invoked again.\n"
                "\n"
                "If you want to cancel automatic reconnect you need to issue the DISCONNECT command.\n"
            ),
        )
        self.commands.register(cmd, self.cmd_connect)

        cmd = CommandParser(
            prog="DISCONNECT",
            description="disconnect from organization",
            epilog=(
                "In addition to disconnecting from an active organization connection this will also cancel any automatic"
                "reconnection attempt.\n"
            ),
        )
        self.commands.register(cmd, self.cmd_disconnect)

        cmd = CommandParser(prog="RECONNECT", description="reconnect to organization")
        self.commands.register(cmd, self.cmd_reconnect)

        cmd = CommandParser(
            prog="SUBSCRIBE",
            description="bridge a stream",
            epilog=(
                "Manually subscribe to a stream and bridge it\n",
                "\n",
                "Any subscriptions will be persisted between reconnects.\n",
            ),
        )
        cmd.add_argument("stream", help="target stream")
        self.commands.register(cmd, self.cmd_subscribe)

        cmd = CommandParser(
            prog="UNSUBSCRIBE",
            description="unbridge a stream and leave the room",
        )
        cmd.add_argument("stream", help="target stream")
        self.commands.register(cmd, self.cmd_unsubscribe)

        cmd = CommandParser(
            prog="SPACE", description="join/create a space for this organization"
        )
        self.commands.register(cmd, self.cmd_space)

        cmd = CommandParser(
            prog="SYNCPERMISSIONS", description="resync all permissions"
        )
        self.commands.register(cmd, self.cmd_syncpermissions)

        cmd = CommandParser(prog="PROFILE", description="fetch our Zulip profile")
        self.commands.register(cmd, self.cmd_profile)

        cmd = CommandParser(
            prog="ROOM",
            description="run a room command from organization room",
            epilog=(
                "Try 'ROOM #foo' to get the list of commands for a room."
                "If a command generates Zulip replies in a bouncer room they will appear in the room itself."
            ),
        )
        cmd.add_argument("target", help="Zulip stream name")
        cmd.add_argument(
            "command", help="Command and arguments", nargs=argparse.REMAINDER
        )
        self.commands.register(cmd, self.cmd_room)

        cmd = CommandParser(
            prog="STATUS", description="show current organization status"
        )
        self.commands.register(cmd, self.cmd_status)

        cmd = CommandParser(
            prog="BACKFILL",
            description="set the default maximum amount of backfilled messages (0 to disable backfilling)",
        )
        cmd.add_argument("amount", nargs="?", help="new amount")
        cmd.add_argument(
            "--update", action="store_true", help="also set this to all existing rooms"
        )
        self.commands.register(cmd, self.cmd_backfill)

        self.mx_register("m.room.message", self.on_mx_message)

    @staticmethod
    async def create(serv: "AppService", organization: dict, user_id: str, name):
        room_id = await serv.create_room(
            name, f"Organization room for {organization['name']}", [user_id]
        )
        room = OrganizationRoom(
            room_id, user_id, serv, [serv.user_id, user_id], bans=[]
        )
        room.from_config(organization)
        serv.register_room(room)

        room.space = await SpaceRoom.create(room, [r.id for r in room.rooms.values()])

        # calls the api and attaches rooms
        await room.space.create_finalize()
        await room.space.attach(room.id)

        await room.save()

        await room.show_help()
        return room

    def from_config(self, config: dict):
        if "name" in config:
            self.name = config["name"]
        else:
            raise Exception("No name key in config for OrganizationRoom")

        if "api_key" in config and config["api_key"]:
            self.api_key = config["api_key"]

        if "email" in config and config["email"]:
            self.email = config["email"]

        if "site" in config and config["site"]:
            self.site = config["site"]

        if "messages" in config and config["messages"]:
            self.messages = config["messages"]

        if "max_backfill_amount" in config and config["max_backfill_amount"]:
            self.max_backfill_amount = config["max_backfill_amount"]

    def to_config(self) -> dict:
        return {
            "name": self.name,
            "api_key": self.api_key,
            "email": self.email,
            "site": self.site,
            "messages": self.messages,
            "max_backfill_amount": self.max_backfill_amount,
        }

    def is_valid(self) -> bool:
        if self.name is None:
            return False

        # we require user to be in organization room or be connected with channels or PMs
        if not self.in_room(self.user_id):
            # if not connected (or trying to) we can clean up
            if not self.connected:
                return False

            # only if post_init has been done and we're connected with no rooms clean up
            if self.post_init_done and self.connected and len(self.rooms) == 0:
                return False

        return True

    def cleanup(self) -> None:
        logging.debug(f"Network {self.id} cleaning up")

        # prevent reconnecting ever again
        self.connected = False
        self.disconnect = True

        if self.caps_task:
            self.caps_task.cancel()
            self.caps_task = None
            logging.debug("... cancelled caps task")

        if self.backoff_task:
            self.backoff_task.cancel()
            self.backoff_task = None
            logging.debug("... cancelled backoff task")

        if self.zulip:
            raise NotImplementedError("Zulip disconnect is not implemented")

        if self.space:
            self.serv.unregister_room(self.space.id)
            self.space.cleanup()
            asyncio.ensure_future(
                self.serv.leave_room(self.space.id, self.space.members)
            )
            logging.debug("... cleaned up space")

        super().cleanup()

    async def show_help(self):
        self.send_notice_html(
            f"Welcome to the organization room for <b>{html.escape(self.name)}</b>!"
        )

        try:
            return await self.commands.trigger("HELP")
        except CommandParserError as e:
            return self.send_notice(str(e))

    async def on_mx_message(self, event) -> None:
        if str(event.content.msgtype) != "m.text" or event.sender == self.serv.user_id:
            return

        # ignore edits
        if event.content.get_edit():
            return

        try:
            lines = event.content.body.split("\n")

            command = lines.pop(0)
            tail = "\n".join(lines) if len(lines) > 0 else None

            await self.commands.trigger(command, tail)
        except CommandParserError as e:
            self.send_notice(str(e))

    async def cmd_connect(self, _args) -> None:
        await self.connect()

    @connected  # pylint: disable=used-before-assignment
    async def cmd_disconnect(self, _args) -> None:
        self.disconnect = True

        if self.backoff_task:
            self.backoff_task.cancel()
            self.backoff_task = None

        if self.caps_task:
            self.caps_task.cancel()
            self.caps_task = None

        self.backoff = 0
        self.next_server = 0
        self.connected_at = 0

        if self.connected:
            self.connected = False
            await self.save()

        if self.zulip:
            self.zulip = None
            self.send_notice("Disconnected")

    async def cmd_reconnect(self, _args) -> None:
        await self.cmd_disconnect(Namespace())
        await self.cmd_connect(Namespace())

    @connected
    async def cmd_subscribe(self, args) -> None:
        stream = args.stream

        for room in self.rooms.values():
            if stream.lower() == room.name:
                self.send_notice(f"Stream {stream} already exists at {room.id}.")
                return

        self.zulip.add_subscriptions([{"name": stream}])
        room = await StreamRoom.create(organization=self, name=stream)
        await room.backfill_messages()

    @connected
    async def cmd_unsubscribe(self, args) -> None:
        stream = args.stream.lower()

        removed_rooms = False

        for room in self.rooms.values():
            if room.name.lower() == stream:
                removed_rooms = True

                self.serv.unregister_room(room.id)
                room.cleanup()
                await self.serv.leave_room(room.id, room.members)
                del self.rooms[room.stream_id]

                self.zulip.remove_subscriptions([stream])
                self.send_notice(
                    f"Unsubscribed from {stream} and removed room {room.id}."
                )

        if not removed_rooms:
            self.send_notice("No room with that name exists.")

    def get_fullname(self):
        if self.fullname:
            return self.fullname
        return self.profile["full_name"]

    async def cmd_fullname(self, args) -> None:
        if args.fullname is None:
            fullname = self.get_fullname()
            if self.zulip and self.zulip.has_connected:
                self.send_notice(f"Full name: {fullname}")
            return

        if self.zulip and self.zulip.has_connected:
            self.zulip.update_user_by_id(
                self.profile["user_id"], full_name=args.fullname
            )
            self.fullname = args.fullname
            self.send_notice(f"Full name set to {self.fullname}")

    async def cmd_site(self, args) -> None:
        if not args.site:
            self.send_notice(f"Zulip site is: {self.site}")
            return

        self.site = args.site
        await self.save()
        self.send_notice(f"Zulip site set to {self.site}")

    async def cmd_email(self, args) -> None:
        if not args.email:
            self.send_notice(f"Bot email is: {self.email}")
            return

        self.email = args.email
        await self.save()
        self.send_notice(f"Bot email set to {self.email}")

    async def cmd_apikey(self, args) -> None:
        if not args.api_key:
            self.send_notice(f"Bot API key is {'not ' if self.api_key else ''}set")
            return

        self.api_key = args.api_key
        await self.save()
        self.send_notice("Bot API Key changed")

    async def cmd_profile(self, _args) -> None:
        self.profile = self.zulip.get_profile()
        self.send_notice(json.dumps(self.profile, indent=4))

    async def cmd_room(self, args) -> None:
        target = args.target.lower()

        room = None

        for r in self.rooms.values():
            if r.name.lower() == target:
                room = r

        if not room:
            self.send_notice(f"No room for {args.target}")
            return

        if len(args.command) == 0:
            args.command = ["HELP"]

        await room.commands.trigger_args(args.command, forward=True)

    async def cmd_status(self, _args) -> None:
        if self.connected_at > 0:
            conntime = asyncio.get_running_loop().time() - self.connected_at
            conntime = str(datetime.timedelta(seconds=int(conntime)))
            self.send_notice(f"Connected for {conntime}")

            if self.real_host[0] != "?":
                self.send_notice(f"Connected from host {self.real_host}")
        else:
            self.send_notice("Not connected to server.")

        pms = []
        chans = []
        plumbs = []

        for room in self.rooms.values():
            match room:
                case PrivateRoom():
                    pms.append(room.name)
                case StreamRoom():
                    chans.append(room.name)

        if len(chans) > 0:
            self.send_notice(f"Channels: {', '.join(chans)}")

        if len(plumbs) > 0:
            self.send_notice(f"Plumbs: {', '.join(plumbs)}")

        if len(pms) > 0:
            self.send_notice(f"PMs: {', '.join(pms)}")

    async def cmd_space(self, _args) -> None:
        if self.space is None:
            # sync create to prevent race conditions
            self.space = SpaceRoom.create(
                self, [room.id for room in self.rooms.values()]
            )

            # calls the api and attaches rooms
            self.send_notice("Creating space and inviting you to it.")
            await self.space.create_finalize()
        else:
            self.send_notice(f"Space already exists ({self.space.id}).")

    async def cmd_syncpermissions(self, _args) -> None:
        await self._sync_permissions()
        self.send_notice("Permissions synched successfully")

    async def cmd_backfill(self, args) -> None:
        if args.amount:
            self.max_backfill_amount = int(args.amount)
            await self.save()
        if args.update:
            for room in self.rooms.values():
                if not isinstance(room, PrivateRoom):
                    continue
                room.max_backfill_amount = self.max_backfill_amount
                await room.save()
            self.send_notice(
                f"Set maximum backfill amount to {self.max_backfill_amount} and updated all rooms"
            )
        else:
            self.send_notice(
                f"Maximum backfill amount is set to: {self.max_backfill_amount}"
            )

    async def connect(self) -> None:
        if not self.is_valid():
            logging.warning(
                "Trying to connect an invalid organization {self.id}, this is likely a dangling organization."
            )
            return

        if self.connlock.locked():
            self.send_notice("Already connecting.")
            return

        async with self.connlock:
            if self.zulip and self.connected:
                self.send_notice("Already connected.")
                return

            self.disconnect = False
            await self._connect()

    async def post_init(self) -> None:
        # attach loose sub-rooms to us
        for room_type in [PrivateRoom, StreamRoom]:
            for room in self.serv.find_rooms(room_type, self.user_id):
                if room.stream_id not in self.rooms and (
                    room.organization_id == self.id
                    or (
                        room.organization_id is None
                        and room.organization_name == self.name
                    )
                ):
                    room.organization = self
                    # this doubles as a migration
                    if room.organization_id is None:
                        logging.debug(f"{self.id} attaching and migrating {room.id}")
                        room.organization_id = self.id
                        await room.save()
                    else:
                        logging.debug(f"{self.id} attaching {room.id}")
                    self.rooms[room.stream_id] = room

        self.post_init_done = True

    async def _connect(self) -> None:
        if not self.site:
            self.send_notice("Zulip site is not set!")
            return
        if not self.email:
            self.send_notice("Bot email is not set!")
            return
        if not self.api_key:
            self.send_notice("Bot API key is not set!")
            return
        # force cleanup
        if self.zulip:
            self.zulip = None

        # reset whois and kickbans buffers
        self.whois_data.clear()
        self.pending_kickbans.clear()

        while not self.disconnect:
            if self.name not in self.serv.config["organizations"]:
                self.send_notice(
                    "This organization does not exist on this bridge anymore."
                )
                return

            try:
                asyncio.ensure_future(
                    self.serv.push_bridge_state(
                        BridgeStateEvent.CONNECTING, remote_id=self.name
                    )
                )

                self.zulip = zulip.Client(
                    self.email, api_key=self.api_key, site=self.site
                )

                if not self.connected:
                    self.connected = True
                    await self.save()

                # awaiting above allows disconnect to happen in-between
                if self.zulip is None:
                    logging.debug("Zulip disconnected")
                    return

                self.disconnect = False
                self.connected_at = asyncio.get_running_loop().time()

                self.profile = self.zulip.get_profile()

                self.zulip_handler = ZulipEventHandler(self)

                # Start Zulip event listerner
                asyncio.get_running_loop().run_in_executor(
                    None,
                    functools.partial(
                        self.zulip.call_on_each_event,
                        lambda event: self.zulip_handler.on_event(  # pylint: disable=unnecessary-lambda
                            event
                        ),
                        apply_markdown=True,
                    ),
                )

                asyncio.ensure_future(self._on_connect())

                self.send_notice(f"Connected to {self.site}")

                asyncio.ensure_future(
                    self.serv.push_bridge_state(
                        BridgeStateEvent.CONNECTED, remote_id=self.name
                    )
                )

                return
            except Exception as e:
                self.send_notice(f"Failed to connect: {str(e)}")

            if self.backoff < 1800:
                self.backoff = self.backoff * 2
            if self.backoff < 5:
                self.backoff = 5

            self.send_notice(f"Retrying in {self.backoff} seconds...")

            self.backoff_task = asyncio.ensure_future(asyncio.sleep(self.backoff))
            try:
                await self.backoff_task
            except asyncio.CancelledError:
                break
            finally:
                self.backoff_task = None

        self.send_notice("Connection aborted.")

    def on_disconnect(self, conn, event) -> None:
        if self.caps_task:
            self.caps_task.cancel()
            self.caps_task = None

        if self.zulip:
            self.zulip = None

        # if we were connected for a while, consider the server working
        if (
            self.connected_at > 0
            and asyncio.get_running_loop().time() - self.connected_at > 300
        ):
            self.backoff = 0
            self.next_server = 0
            self.connected_at = 0

        if self.connected and not self.disconnect:
            if self.backoff < 1800:
                self.backoff += 5

            self.send_notice(f"Disconnected, reconnecting in {self.backoff} seconds...")

            async def later(self):
                self.backoff_task = asyncio.ensure_future(asyncio.sleep(self.backoff))
                try:
                    await self.backoff_task
                    await self.connect()
                except asyncio.CancelledError:
                    self.send_notice("Reconnect cancelled.")
                finally:
                    self.backoff_task = None

            asyncio.ensure_future(later(self))
            asyncio.ensure_future(
                self.serv.push_bridge_state(
                    BridgeStateEvent.TRANSIENT_DISCONNECT, remote_id=self.name
                )
            )
        else:
            self.send_notice("Disconnected.")
            asyncio.ensure_future(
                self.serv.push_bridge_state(
                    BridgeStateEvent.LOGGED_OUT, remote_id=self.name
                )
            )

    def source_text(self, conn, event) -> str:
        source = None

        if event.source is not None:
            source = str(event.source.nick)

            if event.source.user is not None and event.source.host is not None:
                source += f" ({event.source.user}@{event.source.host})"
        else:
            source = conn.server

        return source

    def on_quit(self, conn, event) -> None:
        # leave channels
        for room in self.rooms.values():
            if isinstance(room, StreamRoom):
                room.on_quit(conn, event)

    async def _on_connect(self):
        await self._sync_permissions()
        await self._sync_all_room_members()
        await self.backfill_messages()

    async def _sync_permissions(self):
        # Owner should have the highest permissions (after bot)
        self.permissions[self.serv.config["owner"]] = 99

        # TODO: Get permissions from zulip

        rooms = set(self.rooms.values())
        rooms.add(self)
        rooms.add(self.space)
        logging.info(len(rooms))

        for room in rooms:
            if room is None:
                continue
            logging.debug(f"Synching permissions in {self.name} - {room.id}")
            await room.sync_permissions(self.permissions)

    async def _sync_all_room_members(self):
        result = self.zulip.get_subscriptions(request={"include_subscribers": True})
        if result["result"] != "success":
            logging.error(
                f"Getting subscriptions for {self.name} failed! Message: {result['msg']}"
            )
            return
        streams = result["subscriptions"]
        for stream in streams:
            room = self.rooms.get(stream["stream_id"])
            if not room or not isinstance(room, StreamRoom):
                continue
            asyncio.ensure_future(room.sync_zulip_members(stream["subscribers"]))

    def get_zulip_user(self, user_id: int, update_cache: bool = False):
        if update_cache or user_id not in self.zulip_users:
            result = self.zulip.get_user_by_id(user_id)
            if result["result"] != "success":
                raise Exception(f"Could not get Zulip user {user_id}: {result['msg']}")
            self.zulip_users[user_id] = result["user"]
        return self.zulip_users[user_id]

    def get_zulip_user_id_from_mxid(self, mxid: str) -> str:
        ret = re.search(
            rf"@{self.serv.puppet_prefix}{self.organization.name.lower()}{self.serv.puppet_separator}(\d+):{self.serv.server_name}",
            mxid,
        )
        return ret.group(1)

    async def backfill_messages(self):
        for room in self.rooms.values():
            if room.max_backfill_amount == 0:
                continue
            if isinstance(room, StreamRoom):
                await room.backfill_messages()



# MatrixZulipBridge - an appservice puppeting bridge for Matrix - Zulip
#
# Copyright (C) 2024 Emma Meijere <emgh@em.id.lv>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Originally licensed under the MIT (Expat) license:
# <https://github.com/hifi/heisenbridge/blob/2532905f13835762870de55ba8a404fad6d62d81/LICENSE>.
#
# [This file includes modifications made by Emma Meijere]
#
#
import asyncio
import logging
from typing import TYPE_CHECKING

from mautrix.api import Method, Path
from mautrix.types import SpaceChildStateEventContent
from mautrix.types.event.type import EventType

from matrixzulipbridge.room import Room

if TYPE_CHECKING:
    from matrixzulipbridge.organization_room import OrganizationRoom


class SpaceRoom(Room):
    name: str
    organization: "OrganizationRoom"
    organization_id: str

    # pending rooms to attach during space creation
    pending: list[str]

    def init(self) -> None:
        super().init()

        self.name = None
        self.organization = None
        self.organization_id = None

        self.pending = []

    def is_valid(self) -> bool:
        # we need to know our organization
        if self.organization_id is None:
            return False

        # we are valid as long as our user is in the room
        if not self.in_room(self.user_id):
            return False

        return True

    @staticmethod
    async def create(
        organization: "OrganizationRoom", initial_rooms: list[str]
    ) -> "SpaceRoom":
        logging.debug(
            f"SpaceRoom.create(organization='{organization.id}' ({organization.name}))"
        )

        room = SpaceRoom(
            None,
            organization.user_id,
            organization.serv,
            [organization.user_id, organization.serv.user_id],
            [],
        )
        room.name = organization.name
        room.organization = organization  # only used in create_finalize
        room.organization_id = organization.id
        room.pending += initial_rooms
        return room

    async def create_finalize(self) -> None:
        resp = await self.az.intent.api.request(
            Method.POST,
            Path.v3.createRoom,
            {
                "creation_content": {
                    "type": "m.space",
                },
                "visibility": "private",
                "name": self.organization.name,
                "topic": f"Organization space for {self.organization.name}",
                "invite": [self.organization.user_id],
                "is_direct": False,
                "initial_state": [
                    {
                        "type": "m.space.child",
                        "state_key": self.organization.id,
                        "content": {"via": [self.organization.serv.server_name]},
                    }
                ],
                "power_level_content_override": {
                    "events_default": 50,
                    "users_default": 0,
                    "invite": 50,
                    "kick": 50,
                    "redact": 50,
                    "ban": 50,
                    "events": {
                        "m.room.name": 0,
                        "m.room.avatar": 0,  # these work as long as rooms are private
                    },
                    "users": self.organization.permissions
                    | {self.organization.serv.user_id: 100},
                },
            },
        )

        self.id = resp["room_id"]
        self.serv.register_room(self)
        await self.save()

        # attach all pending rooms
        rooms = self.pending
        self.pending = []

        for room_id in rooms:
            await self.attach(room_id)

    def from_config(self, config: dict) -> None:
        super().from_config(config)

        if "organization_id" in config:
            self.organization_id = config["organization_id"]

    def to_config(self) -> dict:
        return {
            **(super().to_config()),
            "organization_id": self.organization_id,
        }

    def cleanup(self) -> None:
        try:
            organization = self.serv._rooms[self.organization_id]

            if organization.space == self:
                organization.space = None
                organization.space_id = None
                asyncio.ensure_future(organization.save())
                logging.debug(
                    f"Space {self.id} cleaned up from organization {organization.id}"
                )
            else:
                logging.debug(
                    f"Space room cleaned up as a duplicate for organization {organization.id}, probably fine."
                )
        except KeyError:
            logging.debug(
                f"Space room cleaned up with missing organization {self.organization_id}, probably fine."
            )

        super().cleanup()

    async def attach(self, room_id) -> None:
        # if we are attached between space request and creation just add to pending list
        if self.id is None:
            logging.debug(f"Queuing room {room_id} attachment to pending space.")
            self.pending.append(room_id)
            return

        logging.debug(f"Attaching room {room_id} to space {self.id}.")
        await self.az.intent.send_state_event(
            self.id,
            EventType.SPACE_CHILD,  # pylint: disable=no-member
            state_key=room_id,
            content=SpaceChildStateEventContent(via=[self.serv.server_name]),
        )

    async def detach(self, room_id) -> None:
        if self.id is not None:
            logging.debug(f"Detaching room {room_id} from space {self.id}.")
            await self.az.intent.send_state_event(
                self.id,
                EventType.SPACE_CHILD,  # pylint: disable=no-member
                state_key=room_id,
                content=SpaceChildStateEventContent(),
            )
        elif room_id in self.pending:
            logging.debug(f"Removing {room_id} from space {self.id} pending queue.")
            self.pending.remove(room_id)

    async def post_init(self) -> None:
        try:
            organization = self.serv._rooms[self.organization_id]
            if organization.space is not None:
                logging.warning(
                    f"Network room {organization.id} already has space {organization.space.id} but I'm {self.id}, we are dangling."
                )
                return

            organization.space = self
            logging.debug(f"Space {self.id} attached to organization {organization.id}")
        except KeyError:
            logging.debug(
                f"Network room {self.organization_id} was not found for space {self.id}, we are dangling."
            )
            self.organization_id = None



# MatrixZulipBridge - an appservice puppeting bridge for Matrix - Zulip
#
# Copyright (C) 2024 Emma Meijere <emgh@em.id.lv>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Originally licensed under the MIT (Expat) license:
# <https://github.com/hifi/heisenbridge/blob/2532905f13835762870de55ba8a404fad6d62d81/LICENSE>.
#
# [This file includes modifications made by Emma Meijere]
#
#
import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from mautrix.api import Method, Path
from mautrix.errors import MNotFound

if TYPE_CHECKING:
    from mautrix.appservice import AppService as MauService

    from matrixzulipbridge.room import Room


class AppService(ABC):
    az: "MauService"

    user_id: str
    server_name: str
    config: dict

    async def load(self):
        try:
            self.config.update(await self.az.intent.get_account_data("zulip"))
        except MNotFound:
            await self.save()

    async def save(self):
        await self.az.intent.set_account_data("zulip", self.config)

    async def create_room(
        self,
        name: str,
        topic: str,
        invite: list[str],
        restricted: str = None,
        permissions: dict = None,
    ) -> str:
        if permissions is None:
            permissions = {}

        req = {
            "visibility": "private",
            "name": name,
            "topic": topic,
            "invite": invite,
            "is_direct": False,
            "power_level_content_override": {
                "users_default": 0,
                "invite": 50,
                "kick": 50,
                "redact": 50,
                "ban": 50,
                "events": {
                    "m.room.name": 0,
                    "m.room.avatar": 0,  # these work as long as rooms are private
                    "m.room.encryption": 100,
                    "m.space.parent": 90,
                },
                "users": {self.user_id: 100} | permissions,
            },
            "com.beeper.auto_join_invites": True,
        }

        if restricted is not None:
            resp = await self.az.intent.api.request(Method.GET, Path.v3.capabilities)
            try:
                def_ver = resp["capabilities"]["m.room_versions"]["default"]
            except KeyError:
                logging.debug("Unexpected capabilities reply")
                def_ver = None

            # If room version is in range of 1..8, request v9
            if def_ver in [str(v) for v in range(1, 9)]:
                req["room_version"] = "9"

            req["initial_state"] = [
                {
                    "type": "m.room.join_rules",
                    "state_key": "",
                    "content": {
                        "join_rule": "restricted",
                        "allow": [{"type": "m.room_membership", "room_id": restricted}],
                    },
                }
            ]

        resp = await self.az.intent.api.request(Method.POST, Path.v3.createRoom, req)

        return resp["room_id"]

    @abstractmethod
    def register_room(self, room: "Room"):
        pass

    @abstractmethod
    def find_rooms(self, rtype=None, user_id: str = None) -> list["Room"]:
        pass



# MatrixZulipBridge - an appservice puppeting bridge for Matrix - Zulip
#
# Copyright (C) 2024 Emma Meijere <emgh@em.id.lv>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Originally licensed under the MIT (Expat) license:
# <https://github.com/hifi/heisenbridge/blob/2532905f13835762870de55ba8a404fad6d62d81/LICENSE>.
#
# [This file includes modifications made by Emma Meijere]
#
#
import re
from argparse import Namespace
from html import escape
from urllib.parse import urlparse

from mautrix.errors import MatrixRequestError

from matrixzulipbridge import __version__
from matrixzulipbridge.command_parse import (
    CommandManager,
    CommandParser,
    CommandParserError,
)
from matrixzulipbridge.organization_room import OrganizationRoom
from matrixzulipbridge.room import Room


class ControlRoom(Room):
    commands: CommandManager

    def init(self):
        self.commands = CommandManager()

        cmd = CommandParser(
            prog="ORGANIZATIONS", description="list available Zulip organizations"
        )
        self.commands.register(cmd, self.cmd_organizations)

        cmd = CommandParser(prog="OPEN", description="open organization for connecting")
        cmd.add_argument("name", help="organization name (see ORGANIZATIONS)")
        cmd.add_argument(
            "--new",
            action="store_true",
            help="force open a new organization connection",
        )
        self.commands.register(cmd, self.cmd_open)

        cmd = CommandParser(
            prog="STATUS",
            description="show bridge status",
            epilog="Note: admins see all users but only their own rooms",
        )
        self.commands.register(cmd, self.cmd_status)

        cmd = CommandParser(
            prog="QUIT",
            description="disconnect from all organizations",
            epilog=(
                "For quickly leaving all organizations and removing configurations in a single command.\n"
                "\n"
                "Additionally this will close current DM session with the bridge.\n"
            ),
        )
        self.commands.register(cmd, self.cmd_quit)

        if self.serv.is_admin(self.user_id):
            cmd = CommandParser(prog="MASKS", description="list allow masks")
            self.commands.register(cmd, self.cmd_masks)

            cmd = CommandParser(
                prog="ADDMASK",
                description="add new allow mask",
                epilog=(
                    "For anyone else than the owner to use this bridge they need to be allowed to talk with the bridge bot.\n"
                    "This is accomplished by adding an allow mask that determines their permission level when using the bridge.\n"
                    "\n"
                    "Only admins can manage organizations, normal users can just connect.\n"
                ),
            )
            cmd.add_argument(
                "mask", help="Matrix ID mask (eg: @friend:contoso.com or *:contoso.com)"
            )
            cmd.add_argument("--admin", help="Admin level access", action="store_true")
            self.commands.register(cmd, self.cmd_addmask)

            cmd = CommandParser(
                prog="DELMASK",
                description="delete allow mask",
                epilog=(
                    "Note: Removing a mask only prevents starting a new DM with the bridge bot. Use FORGET for ending existing"
                    " sessions."
                ),
            )
            cmd.add_argument(
                "mask", help="Matrix ID mask (eg: @friend:contoso.com or *:contoso.com)"
            )
            self.commands.register(cmd, self.cmd_delmask)

            cmd = CommandParser(
                prog="ADDORGANIZATION", description="add a Zulip organization"
            )
            cmd.add_argument("name", help="server address")
            self.commands.register(cmd, self.cmd_addorganization)

            cmd = CommandParser(
                prog="DELORGANIZATION", description="delete a Zulip organization"
            )
            cmd.add_argument("name", help="organization name")
            self.commands.register(cmd, self.cmd_delorganization)

            cmd = CommandParser(prog="FORGET", description="Forget a Matrix user")
            cmd.add_argument("user", help="Matrix ID (eg: @ex-friend:contoso.com)")
            self.commands.register(cmd, self.cmd_forget)

            cmd = CommandParser(
                prog="DISPLAYNAME", description="change bridge displayname"
            )
            cmd.add_argument("displayname", help="new bridge displayname")
            self.commands.register(cmd, self.cmd_displayname)

            cmd = CommandParser(prog="AVATAR", description="change bridge avatar")
            cmd.add_argument("url", help="new avatar URL (mxc:// format)")
            self.commands.register(cmd, self.cmd_avatar)

            cmd = CommandParser(
                prog="MEDIAURL", description="configure media URL for links"
            )
            cmd.add_argument("url", nargs="?", help="new URL override")
            cmd.add_argument(
                "--remove",
                help="remove URL override (will retry auto-detection)",
                action="store_true",
            )
            self.commands.register(cmd, self.cmd_media_url)

            cmd = CommandParser(
                prog="MEDIAPATH", description="configure media path for links"
            )
            cmd.add_argument("path", nargs="?", help="new path override")
            cmd.add_argument(
                "--remove", help="remove path override", action="store_true"
            )
            self.commands.register(cmd, self.cmd_media_path)

            cmd = CommandParser(prog="VERSION", description="show bridge version")
            self.commands.register(cmd, self.cmd_version)

        self.mx_register("m.room.message", self.on_mx_message)

    def is_valid(self) -> bool:
        if self.user_id is None:
            return False

        if len(self.members) != 2:
            return False

        return True

    async def show_help(self):
        self.send_notice_html(
            f"<b>Howdy, stranger!</b> You have been granted access to the Zulip bridge of <b>{self.serv.server_name}</b>."
        )

        try:
            return await self.commands.trigger("HELP")
        except CommandParserError as e:
            return self.send_notice(str(e))

    async def on_mx_message(self, event) -> bool:
        if str(event.content.msgtype) != "m.text" or event.sender == self.serv.user_id:
            return

        # ignore edits
        if event.content.get_edit():
            return

        try:
            lines = event.content.body.split("\n")

            command = lines.pop(0)
            tail = "\n".join(lines) if len(lines) > 0 else None

            await self.commands.trigger(command, tail)
        except CommandParserError as e:
            self.send_notice(str(e))

    def organizations(self):
        organizations = {}

        for organization, config in self.serv.config["organizations"].items():
            config["name"] = organization
            organizations[organization.lower()] = config

        return organizations

    async def cmd_masks(self, args):
        msg = "Configured masks:\n"

        for mask, value in self.serv.config["allow"].items():
            msg += "\t{} -> {}\n".format(mask, value)

        self.send_notice(msg)

    async def cmd_addmask(self, args):
        masks = self.serv.config["allow"]

        if args.mask in masks:
            return self.send_notice("Mask already exists")

        masks[args.mask] = "admin" if args.admin else "user"
        await self.serv.save()

        self.send_notice("Mask added.")

    async def cmd_delmask(self, args):
        masks = self.serv.config["allow"]

        if args.mask not in masks:
            return self.send_notice("Mask does not exist")

        del masks[args.mask]
        await self.serv.save()

        self.send_notice("Mask removed.")

    async def cmd_organizations(self, args):
        organizations: dict["OrganizationRoom", dict] = self.serv.config[
            "organizations"
        ]

        self.send_notice("Configured organizations:")

        for _, data in organizations.items():
            self.send_notice(f"\t{data}")

    async def cmd_addorganization(self, args):
        organizations = self.organizations()

        if args.name.lower() in organizations:
            return self.send_notice("Organization with that name already exists")

        self.serv.config["organizations"][args.name] = {
            "name": args.name,
        }
        await self.serv.save()

        self.send_notice("Organization added.")

    async def cmd_delorganization(self, args):
        organizations = self.organizations()

        if args.name.lower() not in organizations:
            return self.send_notice("Organization does not exist")

        del self.serv.config["organizations"][args.name.lower()]
        await self.serv.save()

        return self.send_notice("Organization removed.")

    async def cmd_status(self, args):
        users = set()

        if self.serv.is_admin(self.user_id):
            for room in self.serv.find_rooms():
                if not room.user_id:
                    continue
                users.add(room.user_id)

            users = list(users)
            users.sort()
        else:
            users.add(self.user_id)

        self.send_notice_html(f"I have {len(users)} known users:")
        for user_id in users:
            ncontrol = len(self.serv.find_rooms("ControlRoom", user_id))

            self.send_notice_html(
                f"{indent(1)}{user_id} ({ncontrol} open control rooms):"
            )

            for organization in self.serv.find_rooms("OrganizationRoom", user_id):
                connected = "not connected"
                channels = "not in channels"
                privates = "not in PMs"
                plumbs = "not in plumbs"

                if organization.zulip and organization.zulip.has_connected:
                    user = (
                        organization.real_user
                        if organization.real_user[0] != "?"
                        else "?"
                    )
                    host = (
                        organization.real_host
                        if organization.real_host[0] != "?"
                        else "?"
                    )
                    connected = f"connected as {organization.profile['full_name']}!{user}@{host}"

                nchannels = 0
                nprivates = 0
                nplumbs = 0

                for room in organization.rooms.values():
                    if type(room).__name__ == "PrivateRoom":
                        nprivates += 1
                    if type(room).__name__ == "StreamRoom":
                        nchannels += 1

                if nprivates > 0:
                    privates = f"in {nprivates} PMs"

                if nchannels > 0:
                    channels = f"in {nchannels} channels"

                if nplumbs > 0:
                    plumbs = f"in {nplumbs} plumbs"

                self.send_notice_html(
                    f"{indent(2)}{organization.name}, {connected}, {channels}, {privates}, {plumbs}"
                )

                if self.user_id == user_id:
                    for room in organization.rooms.values():
                        join = ""
                        if not room.in_room(user_id):
                            join = " (you have not joined this room)"
                            # ensure the user invite is valid
                            await self.az.intent.invite_user(room.id, self.user_id)
                        self.send_notice_html(
                            f'{indent(3)}<a href="https://matrix.to/#/{escape(room.id)}">{escape(room.name)}</a>{join}'
                        )

    async def cmd_forget(self, args):
        if args.user == self.user_id:
            return self.send_notice("I can't forget you, silly!")

        rooms = self.serv.find_rooms(None, args.user)

        if len(rooms) == 0:
            return self.send_notice("No such user. See STATUS for list of users.")

        # disconnect each organization room in first pass
        for room in rooms:
            if type(room) == OrganizationRoom and room.conn and room.conn.connected:
                self.send_notice(f"Disconnecting {args.user} from {room.name}...")
                await room.cmd_disconnect(Namespace())

        self.send_notice(f"Leaving all {len(rooms)} rooms {args.user} was in...")

        # then just forget everything
        for room in rooms:
            self.serv.unregister_room(room.id)

            try:
                await self.az.intent.leave_room(room.id)
            except MatrixRequestError:
                pass
            try:
                await self.az.intent.forget_room(room.id)
            except MatrixRequestError:
                pass

        self.send_notice(f"Done, I have forgotten about {args.user}")

    async def cmd_displayname(self, args):
        try:
            await self.az.intent.set_displayname(args.displayname)
        except MatrixRequestError as e:
            self.send_notice(f"Failed to set displayname: {str(e)}")

    async def cmd_avatar(self, args):
        try:
            await self.az.intent.set_avatar_url(args.url)
        except MatrixRequestError as e:
            self.send_notice(f"Failed to set avatar: {str(e)}")

    async def cmd_ident(self, args):
        idents = self.serv.config["idents"]

        if args.cmd == "list" or args.cmd is None:
            self.send_notice("Configured custom idents:")
            for mxid, ident in idents.items():
                self.send_notice(f"\t{mxid} -> {ident}")
        elif args.cmd == "set":
            if not re.match(r"^[a-z][-a-z0-9]*$", args.ident):
                self.send_notice(f"Invalid ident string: {args.ident}")
                self.send_notice(
                    "Must be lowercase, start with a letter, can contain dashes, letters and numbers."
                )
            else:
                idents[args.mxid] = args.ident
                self.send_notice(f"Set custom ident for {args.mxid} to {args.ident}")
                await self.serv.save()
        elif args.cmd == "remove":
            if args.mxid in idents:
                del idents[args.mxid]
                self.send_notice(f"Removed custom ident for {args.mxid}")
                await self.serv.save()
            else:
                self.send_notice(f"No custom ident for {args.mxid}")

    async def cmd_sync(self, args):
        if args.lazy:
            self.serv.config["member_sync"] = "lazy"
            await self.serv.save()
        elif args.half:
            self.serv.config["member_sync"] = "half"
            await self.serv.save()
        elif args.full:
            self.serv.config["member_sync"] = "full"
            await self.serv.save()

        self.send_notice(f"Member sync is set to {self.serv.config['member_sync']}")

    async def cmd_media_url(self, args):
        if args.remove:
            self.serv.config["media_url"] = None
            await self.serv.save()
            self.serv.endpoint = await self.serv.detect_public_endpoint()
        elif args.url:
            parsed = urlparse(args.url)
            if (
                parsed.scheme in ["http", "https"]
                and not parsed.params
                and not parsed.query
                and not parsed.fragment
            ):
                self.serv.config["media_url"] = args.url
                await self.serv.save()
                self.serv.endpoint = args.url
            else:
                self.send_notice(f"Invalid media URL format: {args.url}")
                return

        self.send_notice(
            f"Media URL override is set to {self.serv.config['media_url']}"
        )
        self.send_notice(f"Current active media URL: {self.serv.endpoint}")

    async def cmd_media_path(self, args):
        if args.remove:
            self.serv.config["media_path"] = None
            await self.serv.save()
            self.serv.media_path = self.serv.DEFAULT_MEDIA_PATH
        elif args.path:
            self.serv.config["media_path"] = args.path
            await self.serv.save()
            self.serv.media_path = args.path

        self.send_notice(
            f"Media Path override is set to {self.serv.config['media_path']}"
        )
        self.send_notice(f"Current active media path: {self.serv.media_path}")

    async def cmd_open(self, args):
        organizations = self.organizations()
        name = args.name.lower()

        if name not in organizations:
            return self.send_notice("Organization does not exist")

        organization = organizations[name]

        found = 0
        for room in self.serv.find_rooms(OrganizationRoom, self.user_id):
            if room.name == organization["name"]:
                found += 1

                if not args.new:
                    if self.user_id not in room.members:
                        self.send_notice(f"Inviting back to {room.name} ({room.id})")
                        await self.az.intent.invite_user(room.id, self.user_id)
                    else:
                        self.send_notice(f"You are already in {room.name} ({room.id})")

        # if we found at least one organization room, no need to create unless forced
        if found > 0 and not args.new:
            return

        name = (
            organization["name"]
            if found == 0
            else f"{organization['name']} {found + 1}"
        )

        self.send_notice(f"You have been invited to {name}")
        await OrganizationRoom.create(self.serv, organization, self.user_id, name)

    async def cmd_quit(self, args):
        rooms = self.serv.find_rooms(None, self.user_id)

        # disconnect each organization room in first pass
        for room in rooms:
            if (
                type(room) == OrganizationRoom
                and room.zulip
                and room.zulip.has_connected
            ):
                self.send_notice(f"Disconnecting from {room.name}...")
                await room.cmd_disconnect(Namespace())

        self.send_notice("Closing all channels and private messages...")

        # then just forget everything
        for room in rooms:
            if room.id == self.id:
                continue

            self.serv.unregister_room(room.id)

            try:
                await self.az.intent.leave_room(room.id)
            except MatrixRequestError:
                pass
            try:
                await self.az.intent.forget_room(room.id)
            except MatrixRequestError:
                pass

    async def cmd_version(self, args):
        self.send_notice(f"zulipbridge v{__version__}")



# MatrixZulipBridge - an appservice puppeting bridge for Matrix - Zulip
#
# Copyright (C) 2024 Emma Meijere <emgh@em.id.lv>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Originally licensed under the MIT (Expat) license:
# <https://github.com/hifi/heisenbridge/blob/2532905f13835762870de55ba8a404fad6d62d81/LICENSE>.
#
# [This file includes modifications made by Emma Meijere]
#
#
import argparse
import asyncio
import grp
import logging
import os
import pwd
import random
import re
import string
import sys
import urllib
from fnmatch import fnmatch
from typing import Optional

from mautrix.api import HTTPAPI, Method, Path, SynapseAdminPath
from mautrix.appservice import AppService as MauService
from mautrix.appservice.state_store import ASStateStore
from mautrix.client.state_store.memory import MemoryStateStore
from mautrix.errors import (
    MatrixConnectionError,
    MatrixRequestError,
    MForbidden,
    MNotFound,
    MUserInUse,
)
from mautrix.types import Membership
from mautrix.util.bridge_state import BridgeState, BridgeStateEvent
from mautrix.util.config import yaml

from matrixzulipbridge import __version__
from matrixzulipbridge.appservice import AppService
from matrixzulipbridge.control_room import ControlRoom
from matrixzulipbridge.organization_room import OrganizationRoom
from matrixzulipbridge.personal_room import PersonalRoom
from matrixzulipbridge.private_room import PrivateRoom
from matrixzulipbridge.room import Room, RoomInvalidError
from matrixzulipbridge.space_room import SpaceRoom
from matrixzulipbridge.stream_room import StreamRoom
from matrixzulipbridge.websocket import AppserviceWebsocket

try:  # Optionally load coloredlogs
    import coloredlogs
except ModuleNotFoundError:
    pass


class MemoryBridgeStateStore(ASStateStore, MemoryStateStore):
    def __init__(self) -> None:
        ASStateStore.__init__(self)
        MemoryStateStore.__init__(self)


class BridgeAppService(AppService):
    _api: HTTPAPI
    _rooms: dict[str, Room]
    _users: dict[str, str]

    DEFAULT_MEDIA_PATH = "/_matrix/media/v3/download/{netloc}{path}{filename}"

    registration: Optional[dict]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.registration = None
        self.puppet_separator = None
        self.puppet_prefix = None
        self.api = None
        self.synapse_admin = None
        self.endpoint = None

    async def push_bridge_state(
        self,
        state_event: BridgeStateEvent,
        error=None,
        message=None,
        ttl=None,
        remote_id=None,
    ) -> None:
        if (
            "zulipbridge" not in self.registration
            or "status_endpoint" not in self.registration["zulipbridge"]
        ):
            return

        state = BridgeState(
            state_event=state_event,
            error=error,
            message=message,
            ttl=ttl,
            remote_id=remote_id,
        )

        logging.debug(f"Updating bridge state {state}")

        await state.send(
            self.registration["zulipbridge"]["status_endpoint"],
            self.az.as_token,
            log=logging,
        )

    def register_room(self, room: Room):
        self._rooms[room.id] = room

    def unregister_room(self, room_id):
        if room_id in self._rooms:
            del self._rooms[room_id]

    # this is mostly used by organization rooms at init, it's a bit slow
    def find_rooms(self, rtype=None, user_id=None) -> list[Room]:
        ret = []

        if rtype is not None and not isinstance(rtype, str):
            rtype = rtype.__name__

        for room in self._rooms.values():
            if (rtype is None or room.__class__.__name__ == rtype) and (
                user_id is None or room.user_id == user_id
            ):
                ret.append(room)

        return ret

    def is_admin(self, user_id: str):
        if user_id == self.config["owner"]:
            return True

        for mask, value in self.config["allow"].items():
            if fnmatch(user_id, mask) and value == "admin":
                return True

        return False

    def is_user(self, user_id: str):
        if self.is_admin(user_id):
            return True

        for mask in self.config["allow"].keys():
            if fnmatch(user_id, mask):
                return True

        return False

    def is_local(self, mxid: str):
        return mxid.endswith(":" + self.server_name)

    def is_puppet(self, mxid: str) -> bool:
        """Checks whether a given MXID is our puppet

        Args:
            mxid (str): Matrix user ID

        Returns:
            bool:
        """
        return mxid.startswith("@" + self.puppet_prefix) and self.is_local(mxid)

    def get_mxid_from_zulip_user_id(
        self, organization, zulip_user_id, at=True, server=True
    ):
        ret = re.sub(
            r"[^0-9a-z\-\.=\_/]",
            lambda m: "=" + m.group(0).encode("utf-8").hex(),
            f"{self.puppet_prefix}{organization.name}{self.puppet_separator}{zulip_user_id}".lower(),
        )
        # ret = f"{self.puppet_prefix}{organization.site}{self.puppet_separator}{zulip_user_id}".lower()

        if at:
            ret = "@" + ret

        if server:
            ret += ":" + self.server_name

        return ret

    async def cache_user(self, user_id, displayname):
        # start by caching that the user_id exists without a displayname
        if user_id not in self._users:
            self._users[user_id] = None

        # if the cached displayname is incorrect
        if displayname and self._users[user_id] != displayname:
            try:
                await self.az.intent.user(user_id).set_displayname(displayname)
                self._users[user_id] = displayname
            except MatrixRequestError as e:
                logging.warning(
                    f"Failed to set displayname '{displayname}' for user_id '{user_id}', got '{e}'"
                )

    def is_user_cached(self, user_id, displayname=None):
        return user_id in self._users and (
            displayname is None or self._users[user_id] == displayname
        )

    async def ensure_zulip_user_id(
        self,
        organization: "OrganizationRoom",
        zulip_user_id=None,
        update_cache=True,
        zulip_user=None,
    ):
        if zulip_user_id is None:
            zulip_user_id = zulip_user["user_id"]

        mx_user_id = self.get_mxid_from_zulip_user_id(organization, zulip_user_id)

        # if we've seen this user before, we can skip registering
        if not self.is_user_cached(mx_user_id):
            await self.az.intent.user(mx_user_id).ensure_registered()

        # always ensure the displayname is up-to-date
        if update_cache:
            zulip_user = organization.get_zulip_user(zulip_user_id)
            await self.cache_user(mx_user_id, zulip_user["full_name"])

        return mx_user_id

    async def _on_mx_event(self, event):
        if event.room_id and event.room_id in self._rooms:
            try:
                room = self._rooms[event.room_id]
                await room.on_mx_event(event)
            except RoomInvalidError:
                logging.info(
                    f"Event handler for {event.type} threw RoomInvalidError, leaving and cleaning up."
                )
                self.unregister_room(room.id)
                room.cleanup()

                await self.leave_room(room.id, room.members)
            except Exception:
                logging.exception(
                    "Ignoring exception from room handler. This should be fixed."
                )
        elif (
            str(event.type) == "m.room.member"
            and event.sender != self.user_id
            and event.content.membership == Membership.INVITE
        ):
            # set owner if we have none and the user is from the same HS
            if self.config.get("owner", None) is None and event.sender.endswith(
                ":" + self.server_name
            ):
                logging.info(f"We have an owner now, let us rejoice, {event.sender}!")
                self.config["owner"] = event.sender
                await self.save()

            if not self.is_user(event.sender):
                logging.info(
                    f"Non-whitelisted user {event.sender} tried to invite us, ignoring."
                )
                return
            else:
                logging.info(f"Got an invite from {event.sender}")

            if not event.content.is_direct:
                logging.debug("Got an invite to non-direct room, ignoring")
                return

            # only respond to invites unknown new rooms
            if event.room_id in self._rooms:
                logging.debug("Got an invite to room we're already in, ignoring")
                return

            # handle invites against puppets
            if event.state_key != self.user_id:
                logging.info(
                    f"Whitelisted user {event.sender} invited {event.state_key}, going to reject."
                )

                try:
                    await self.az.intent.user(event.state_key).kick_user(
                        event.room_id,
                        event.state_key,
                        "Will invite YOU instead",
                    )
                except Exception:
                    logging.exception("Failed to reject invitation.")

                raise NotImplementedError(
                    "Puppet invites as profile query"
                )  #:TODO: implement

                for room in self.find_rooms(OrganizationRoom, event.sender):
                    pass

                return

            logging.info(
                f"Whitelisted user {event.sender} invited us, going to accept."
            )

            # accept invite sequence
            try:
                room = ControlRoom(
                    id=event.room_id,
                    user_id=event.sender,
                    serv=self,
                    members=[event.sender],
                    bans=[],
                )
                await room.save()
                self.register_room(room)

                await self.az.intent.join_room(room.id)

                # show help on open
                await room.show_help()
            except Exception:
                if event.room_id in self._rooms:
                    del self._rooms[event.room_id]
                logging.exception("Failed to create control room.")
        else:
            pass

    async def detect_public_endpoint(self):
        async with self.api.session as session:
            # first try https well-known
            try:
                resp = await session.request(
                    "GET",
                    f"https://{self.server_name}/.well-known/matrix/client",
                )
                data = await resp.json(content_type=None)
                return data["m.homeserver"]["base_url"]
            except Exception:
                logging.debug("Did not find .well-known for HS")

            # try https directly
            try:
                resp = await session.request(
                    "GET", f"https://{self.server_name}/_matrix/client/versions"
                )
                await resp.json(content_type=None)
                return f"https://{self.server_name}"
            except Exception:
                logging.debug("Could not use direct connection to HS")

            # give up
            logging.warning(
                "Using internal URL for homeserver, media links are likely broken!"
            )
            return str(self.api.base_url)

    def mxc_to_url(self, mxc, filename=None):
        mxc = urllib.parse.urlparse(mxc)

        if filename is None:
            filename = ""
        else:
            filename = "/" + urllib.parse.quote(filename)

        media_path = self.media_path.format(
            netloc=mxc.netloc, path=mxc.path, filename=filename
        )

        return self.endpoint + media_path

    async def reset(self, config_file, homeserver_url):
        with open(config_file, encoding="utf-8") as f:
            registration = yaml.load(f)

        api = HTTPAPI(base_url=homeserver_url, token=registration["as_token"])
        whoami = await api.request(Method.GET, Path.v3.account.whoami)
        self.user_id = whoami["user_id"]
        self.server_name = self.user_id.split(":", 1)[1]
        logging.info("We are " + whoami["user_id"])

        self.az = MauService(
            id=registration["id"],
            domain=self.server_name,
            server=homeserver_url,
            as_token=registration["as_token"],
            hs_token=registration["hs_token"],
            bot_localpart=registration["sender_localpart"],
            state_store=MemoryBridgeStateStore(),
        )

        try:
            await self.az.start(host="127.0.0.1", port=None)
        except Exception:
            logging.exception("Failed to listen.")
            return

        joined_rooms = await self.az.intent.get_joined_rooms()
        logging.info(f"Leaving from {len(joined_rooms)} rooms...")

        for room_id in joined_rooms:
            logging.info(f"Leaving from {room_id}...")
            await self.leave_room(room_id, None)

        logging.info("Resetting configuration...")
        self.config = {}
        await self.save()

        logging.info("All done!")

    def load_reg(self, config_file):
        with open(config_file, encoding="utf-8") as f:
            self.registration = yaml.load(f)

    async def leave_room(self, room_id, members):
        members = members if members else []

        for member in members:
            (name, server) = member.split(":", 1)

            if name.startswith("@" + self.puppet_prefix) and server == self.server_name:
                try:
                    await self.az.intent.user(member).leave_room(room_id)
                except Exception:
                    logging.exception("Removing puppet on leave failed")

        try:
            await self.az.intent.leave_room(room_id)
        except MatrixRequestError:
            pass
        try:
            await self.az.intent.forget_room(room_id)
        except MatrixRequestError:
            pass

    def _keepalive(self):
        async def put_presence():
            try:
                await self.az.intent.set_presence(self.user_id)
            except Exception:
                pass

        asyncio.ensure_future(put_presence())
        asyncio.get_running_loop().call_later(60, self._keepalive)

    async def run(
        self, listen_address, listen_port, homeserver_url, owner, unsafe_mode
    ):
        if "sender_localpart" not in self.registration:
            logging.critical("Missing sender_localpart from registration file.")
            sys.exit(1)

        if (
            "namespaces" not in self.registration
            or "users" not in self.registration["namespaces"]
        ):
            logging.critical("User namespaces missing from registration file.")
            sys.exit(1)

        # remove self namespace if exists
        ns_users = [
            x
            for x in self.registration["namespaces"]["users"]
            if x["regex"].split(":")[0] != f"@{self.registration['sender_localpart']}"
        ]

        if len(ns_users) != 1:
            logging.critical(
                "A single user namespace is required for puppets in the registration file."
            )
            sys.exit(1)

        if "exclusive" not in ns_users[0] or not ns_users[0]["exclusive"]:
            logging.critical("User namespace must be exclusive.")
            sys.exit(1)

        m = re.match(r"^@(.+)([\_/])\.[\*\+]:?", ns_users[0]["regex"])
        if not m:
            logging.critical(
                "User namespace regex must be an exact prefix like '@zulip_bridge_.*' that includes the separator character (_ or /)."
            )
            sys.exit(1)

        self.puppet_separator = m.group(2)
        self.puppet_prefix = m.group(1) + self.puppet_separator

        logging.info(f"zulipbridge v{__version__}")
        if unsafe_mode:
            logging.warning("Running in unsafe mode, bridge may leave rooms on error")

        url = urllib.parse.urlparse(homeserver_url)
        ws = None
        if url.scheme in ["ws", "wss"]:
            logging.info(
                f"Using websockets to receive transactions. Listening is still enabled on http://{listen_address}:{listen_port}"
            )
            ws = AppserviceWebsocket(
                homeserver_url, self.registration["as_token"], self._on_mx_event
            )
            homeserver_url = url._replace(
                scheme=("https" if url.scheme == "wss" else "http")
            ).geturl()
            logging.info(f"Connecting to HS at {homeserver_url}")

        self.api = HTTPAPI(base_url=homeserver_url, token=self.registration["as_token"])

        # conduit requires that the appservice user is registered before whoami
        wait = 0
        while True:
            try:
                await self.api.request(
                    Method.POST,
                    Path.v3.register,
                    {
                        "type": "m.login.application_service",
                        "username": self.registration["sender_localpart"],
                    },
                )
                logging.debug("Appservice user registration succeeded.")
                break
            except MUserInUse:
                logging.debug("Appservice user is already registered.")
                break
            except MatrixConnectionError as e:
                if wait < 30:
                    wait += 5
                logging.warning(
                    f"Failed to connect to HS: {e}, retrying in {wait} seconds..."
                )
                await asyncio.sleep(wait)
            except Exception:
                logging.exception(
                    "Unexpected failure when registering appservice user."
                )
                sys.exit(1)

        # mautrix migration requires us to call whoami manually at this point
        whoami = await self.api.request(Method.GET, Path.v3.account.whoami)

        logging.info("We are %s", whoami["user_id"])

        self.user_id = whoami["user_id"]
        self.server_name = self.user_id.split(":", 1)[1]

        self.az = MauService(
            id=self.registration["id"],
            domain=self.server_name,
            server=homeserver_url,
            as_token=self.registration["as_token"],
            hs_token=self.registration["hs_token"],
            bot_localpart=self.registration["sender_localpart"],
            state_store=MemoryBridgeStateStore(),
        )
        self.az.matrix_event_handler(self._on_mx_event)

        try:
            await self.az.start(host=listen_address, port=listen_port)
        except Exception:
            logging.exception("Failed to listen.")
            sys.exit(1)

        try:
            await self.az.intent.ensure_registered()
            logging.debug("Appservice user exists at least now.")
        except Exception:
            logging.exception("Unexpected failure when registering appservice user.")
            sys.exit(1)

        if (
            "zulipbridge" in self.registration
            and "displayname" in self.registration["zulipbridge"]
        ):
            try:
                logging.debug(
                    f"Overriding displayname from registration file to {self.registration['zulipbridge']['displayname']}"
                )
                await self.az.intent.set_displayname(
                    self.registration["zulipbridge"]["displayname"]
                )
            except MatrixRequestError as e:
                logging.warning(f"Failed to set displayname: {str(e)}")

        self._rooms = {}
        self._users = {}
        self.config = {
            "organizations": {},
            "owner": None,
            "member_sync": "half",
            "media_url": None,
            "media_path": None,
            "namespace": self.puppet_prefix,
        }
        logging.debug(f"Default config: {self.config}")
        self.synapse_admin = False

        try:
            is_admin = await self.api.request(
                Method.GET, SynapseAdminPath.v1.users[self.user_id].admin
            )
            self.synapse_admin = is_admin["admin"]
        except MForbidden:
            logging.info(
                f"We ({self.user_id}) are not a server admin, inviting puppets is required."
            )
        except Exception:
            logging.info(
                "Seems we are not connected to Synapse, inviting puppets is required."
            )

        # load config from HS
        await self.load()

        async def _resolve_media_endpoint():
            endpoint = await self.detect_public_endpoint()

            # only rewrite it if it wasn't changed
            if self.endpoint == str(self.api.base_url):
                self.endpoint = endpoint

            logging.info("Homeserver is publicly available at " + self.endpoint)

        # use configured media_url for endpoint if we have it
        if (
            "zulipbridge" in self.registration
            and "media_url" in self.registration["zulipbridge"]
        ):
            logging.debug(
                f"Overriding media URL from registration file to {self.registration['zulipbridge']['media_url']}"
            )
            self.endpoint = self.registration["zulipbridge"]["media_url"]
        elif self.config["media_url"]:
            self.endpoint = self.config["media_url"]
        else:
            logging.info(
                "Trying to detect homeserver public endpoint, this might take a while..."
            )
            self.endpoint = str(self.api.base_url)
            asyncio.ensure_future(_resolve_media_endpoint())

        # use configured media_path for media_path if we have it
        if (
            "zulipbridge" in self.registration
            and "media_path" in self.registration["zulipbridge"]
        ):
            logging.debug(
                f"Overriding media path from registration file to {self.registration['zulipbridge']['media_path']}"
            )
            self.media_path = self.registration["zulipbridge"]["media_path"]
        elif self.config["media_path"]:
            self.media_path = self.config["media_path"]
        else:
            self.media_path = self.DEFAULT_MEDIA_PATH

        logging.info("Starting presence loop")
        self._keepalive()

        # prevent starting bridge with changed namespace
        if self.config["namespace"] != self.puppet_prefix:
            logging.error(
                f"Previously used namespace '{self.config['namespace']}' does not match current '{self.puppet_prefix}'."
            )
            sys.exit(1)

        # honor command line owner
        if owner is not None and self.config["owner"] != owner:
            logging.info(f"Overriding loaded owner with '{owner}'")
            self.config["owner"] = owner

        # always ensure our merged and migrated configuration is up-to-date
        await self.save()

        logging.info("Fetching joined rooms...")

        joined_rooms = await self.az.intent.get_joined_rooms()
        logging.debug(f"Appservice rooms: {joined_rooms}")

        logging.info(f"Bridge is in {len(joined_rooms)} rooms, initializing them...")

        Room.init_class(self.az)

        # room types and their init order, organization must be before chat and group
        room_types = [
            ControlRoom,
            OrganizationRoom,
            PrivateRoom,
            StreamRoom,
            PersonalRoom,
            SpaceRoom,
        ]

        room_type_map = {}
        for room_type in room_types:
            room_type.init_class(self.az)
            room_type_map[room_type.__name__] = room_type

        # we always auto-open control room for owner
        owner_control_open = False

        # import all rooms
        for room_id in joined_rooms:
            joined = {}

            try:
                config = await self.az.intent.get_account_data("zulip", room_id)

                if "type" not in config or "user_id" not in config:
                    raise Exception("Invalid config")

                cls = room_type_map.get(config["type"])
                if not cls:
                    raise Exception("Unknown room type")

                # refresh room members state
                await self.az.intent.get_room_members(room_id)

                joined = await self.az.state_store.get_member_profiles(
                    room_id, (Membership.JOIN,)
                )
                banned = await self.az.state_store.get_members(
                    room_id, (Membership.BAN,)
                )

                room = cls(
                    id=room_id,
                    user_id=config["user_id"],
                    serv=self,
                    members=joined.keys(),
                    bans=banned,
                )
                room.from_config(config)

                # add to room displayname
                for user_id, member in joined.items():
                    if member.displayname is not None:
                        room.displaynames[user_id] = member.displayname
                    # add to global puppet cache if it's a puppet
                    if self.is_puppet(user_id):
                        self._users[user_id] = member.displayname

                # only add valid rooms to event handler
                if room.is_valid():
                    self._rooms[room_id] = room
                else:
                    room.cleanup()
                    raise Exception("Room validation failed after init")

                if cls == ControlRoom and room.user_id == self.config["owner"]:
                    owner_control_open = True
            except MNotFound:
                logging.error(
                    f"Leaving room with no data: {room_id}. How did this happen?"
                )
                self.unregister_room(room_id)
                await self.leave_room(room_id, joined.keys())
            except Exception:
                logging.exception(
                    f"Failed to reconfigure room {room_id} during init, leaving."
                )

                # regardless of safe mode, we ignore this room
                self.unregister_room(room_id)

                if unsafe_mode:
                    await self.leave_room(room_id, joined.keys())

        logging.info("All valid rooms initialized, connecting organization rooms...")

        wait = 1
        for room in list(self._rooms.values()):
            await room.post_init()

            # check again if we're still valid
            if not room.is_valid():
                logging.debug(
                    f"Room {room.id} failed validation after post init, leaving."
                )

                self.unregister_room(room.id)

                if unsafe_mode:
                    await self.leave_room(room.id, room.members)

                continue

            # connect organization rooms one by one, this may take a while
            if isinstance(room, OrganizationRoom) and not room.connected:

                def sync_connect(room):
                    asyncio.ensure_future(room.connect())

                asyncio.get_running_loop().call_later(wait, sync_connect, room)
                wait += 1

        logging.info(
            f"Init done with {wait-1} organizations connecting, bridge is now running!"
        )

        await self.push_bridge_state(BridgeStateEvent.UNCONFIGURED)

        # late start WS to avoid getting transactions too early
        if ws:
            await ws.start()

        if self.config["owner"] and not owner_control_open:
            logging.info(f"Opening control room for owner {self.config['owner']}")
            try:
                room_id = await self.az.intent.create_room(
                    invitees=[self.config["owner"]],
                    custom_request_fields={"com.beeper.auto_join_invites": True},
                )

                room = ControlRoom(
                    id=room_id,
                    user_id=self.config["owner"],
                    serv=self,
                    members=[self.config["owner"]],
                    bans=[],
                )
                await room.save()
                self.register_room(room)

                await self.az.intent.join_room(room.id)

                # show help on open
                await room.show_help()
            except Exception:
                logging.error("Failed to create control room, huh")

        await asyncio.Event().wait()


async def async_main():
    parser = argparse.ArgumentParser(
        prog=os.path.basename(sys.executable) + " -m " + __package__,
        description=f"A puppeting Matrix - Zulip appservice bridge (v{__version__})",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="log debug messages",
        action="store_true",
        default=argparse.SUPPRESS,
    )
    req = parser.add_mutually_exclusive_group(required=True)
    req.add_argument(
        "-c",
        "--config",
        help="registration YAML file path, must be writable if generating",
    )
    req.add_argument(
        "--version",
        action="store_true",
        help="show bridge version",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "-l",
        "--listen-address",
        help="bridge listen address (default: as specified in url in config, 127.0.0.1 otherwise)",
    )
    parser.add_argument(
        "-p",
        "--listen-port",
        help="bridge listen port (default: as specified in url in config, 28464 otherwise)",
        type=int,
    )
    parser.add_argument("-u", "--uid", help="user id to run as", default=None)
    parser.add_argument("-g", "--gid", help="group id to run as", default=None)
    parser.add_argument(
        "--generate",
        action="store_true",
        help="generate registration YAML for Matrix homeserver (Synapse)",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--generate-compat",
        action="store_true",
        help="generate registration YAML for Matrix homeserver (Dendrite and Conduit)",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--reset",
        action="store_true",
        help="reset ALL bridge configuration from homeserver and exit",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--unsafe-mode",
        action="store_true",
        help="allow appservice to leave rooms on error",
    )
    parser.add_argument(
        "-o",
        "--owner",
        help="set owner MXID (eg: @user:homeserver) or first talking local user will claim the bridge",
        default=None,
    )
    parser.add_argument(
        "homeserver",
        nargs="?",
        help="URL of Matrix homeserver",
        default="http://localhost:8008",
    )

    args = parser.parse_args()

    logging_level = logging.INFO
    if "verbose" in args:
        logging_level = logging.DEBUG

    logging.basicConfig(stream=sys.stdout, level=logging_level)

    try:
        coloredlogs.install(logging_level)
    except NameError:
        pass

    if "generate" in args or "generate_compat" in args:
        letters = string.ascii_letters + string.digits

        registration = {
            "id": "zulipbridge",
            "url": f"http://{args.listen_address or '127.0.0.1'}:{args.listen_port or 28464}",
            "as_token": "".join(random.choice(letters) for i in range(64)),
            "hs_token": "".join(random.choice(letters) for i in range(64)),
            "rate_limited": False,
            "sender_localpart": "zulipbridge",
            "namespaces": {
                "users": [{"regex": "@zulip_.*", "exclusive": True}],
                "aliases": [],
                "rooms": [],
            },
        }

        if "generate_compat" in args:
            registration["namespaces"]["users"].append(
                {"regex": "@zulipbridge:.*", "exclusive": True}
            )

        if os.path.isfile(args.config):
            logging.critical("Registration file already exists, not overwriting.")
            sys.exit(1)

        if args.config == "-":
            yaml.dump(registration, sys.stdout)
        else:
            with open(args.config, "w", encoding="utf-8") as f:
                yaml.dump(registration, f)

            logging.info(f"Registration file generated and saved to {args.config}")
    elif "reset" in args:
        service = BridgeAppService()
        logging.warning("Resetting will delete all bridge data, this is irreversible!")
        await asyncio.sleep(3)  # Gotta be careful
        if input("Are you SURE you want to continue? [y/n] ").lower() == "y":
            await service.reset(args.config, args.homeserver)
        else:
            logging.info("Not doing anything.")
            sys.exit(0)
    elif "version" in args:
        logging.info(__version__)
    else:
        service = BridgeAppService()

        service.load_reg(args.config)

        if os.getuid() == 0:
            if args.gid:
                gid = grp.getgrnam(args.gid).gr_gid
                os.setgid(gid)
                os.setgroups([])

            if args.uid:
                uid = pwd.getpwnam(args.uid).pw_uid
                os.setuid(uid)

        os.umask(0o077)

        listen_address = args.listen_address
        listen_port = args.listen_port

        if not listen_address:
            listen_address = "127.0.0.1"

            try:
                url = urllib.parse.urlparse(service.registration["url"])
                if url.hostname:
                    listen_address = url.hostname
            except Exception:
                pass

        if not listen_port:
            listen_port = 28464

            try:
                url = urllib.parse.urlparse(service.registration["url"])
                if url.port:
                    listen_port = url.port
            except Exception:
                pass

        await service.run(
            listen_address, listen_port, args.homeserver, args.owner, args.unsafe_mode
        )


def main():
    asyncio.run(async_main())


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        pass



# MatrixZulipBridge - an appservice puppeting bridge for Matrix - Zulip
#
# Copyright (C) 2024 Emma Meijere <emgh@em.id.lv>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Originally licensed under the MIT (Expat) license:
# <https://github.com/hifi/heisenbridge/blob/2532905f13835762870de55ba8a404fad6d62d81/LICENSE>.
#
# [This file includes modifications made by Emma Meijere]
#
#
import asyncio
import logging
import re
from typing import TYPE_CHECKING, Optional

from mautrix.types import MessageType

from matrixzulipbridge.command_parse import CommandParser
from matrixzulipbridge.private_room import PrivateRoom

if TYPE_CHECKING:
    from matrixzulipbridge.organization_room import OrganizationRoom


def connected(f):
    """Wait until organization has connected to Zulip."""

    def wrapper(*args, **kwargs):
        self = args[0]

        if (
            not self.organization
            or not self.organization.zulip
            or not self.organization.zulip.has_connected
        ):
            return asyncio.sleep(0)

        return f(*args, **kwargs)

    return wrapper


class StreamRoom(PrivateRoom):
    """Puppeting room for Zulip stream."""

    key: Optional[str]
    member_sync: str
    names_buffer: list[str]

    use_displaynames = True
    allow_notice = False
    force_forward = True
    topic_sync = None

    stream_id: Optional[int]
    stream_name: Optional[str]

    def init(self) -> None:
        super().init()

        self.key = None
        self.autocmd = None

        self.stream_id = None
        self.stream_name = None

        # for migration the class default is full
        self.member_sync = "full"

        cmd = CommandParser(
            prog="SYNC",
            description="override Zulip member sync type for this room",
            epilog="Note: To force full sync after setting to full, use the NAMES command",
        )
        group = cmd.add_mutually_exclusive_group()
        group.add_argument(
            "--lazy",
            help="set lazy sync, members are added when they talk",
            action="store_true",
        )
        group.add_argument(
            "--half",
            help="set half sync, members are added when they join or talk",
            action="store_true",
        )
        group.add_argument(
            "--full",
            help="set full sync, members are fully synchronized",
            action="store_true",
        )
        group.add_argument(
            "--off",
            help="disable member sync completely, the bridge will relay all messages, may be useful during spam attacks",
            action="store_true",
        )
        self.commands.register(cmd, self.cmd_sync)

        cmd = CommandParser(
            prog="UPGRADE",
            description="Perform any potential bridge-side upgrades of the room",
        )
        cmd.add_argument(
            "--undo", action="store_true", help="undo previously performed upgrade"
        )
        self.commands.register(cmd, self.cmd_upgrade)

        cmd = CommandParser(
            prog="DISPLAYNAMES",
            description="enable or disable use of displaynames in relayed messages",
        )
        cmd.add_argument(
            "--enable", dest="enabled", action="store_true", help="Enable displaynames"
        )
        cmd.add_argument(
            "--disable",
            dest="enabled",
            action="store_false",
            help="Disable displaynames (fallback to MXID)",
        )
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_displaynames)

        cmd = CommandParser(
            prog="NOTICERELAY",
            description="enable or disable relaying of Matrix notices to Zulip",
        )
        cmd.add_argument(
            "--enable", dest="enabled", action="store_true", help="Enable notice relay"
        )
        cmd.add_argument(
            "--disable",
            dest="enabled",
            action="store_false",
            help="Disable notice relay",
        )
        cmd.set_defaults(enabled=None)
        self.commands.register(cmd, self.cmd_noticerelay)

        cmd = CommandParser(
            prog="TOPIC",
            description="show or set channel topic and configure sync mode",
        )
        cmd.add_argument(
            "--sync",
            choices=["off", "zulip", "matrix", "any"],
            help="Topic sync targets, defaults to off",
        )
        cmd.add_argument("text", nargs="*", help="topic text if setting")
        self.commands.register(cmd, self.cmd_topic)

        self.mx_register("m.room.topic", self._on_mx_room_topic)

    def is_valid(self) -> bool:
        # we are valid as long as the appservice is in the room
        if not self.in_room(self.serv.user_id):
            return False

        return True

    @staticmethod
    async def create(organization: "OrganizationRoom", name: str) -> "StreamRoom":
        logging.debug(
            f"StreamRoom.create(organization='{organization.name}', name='{name}'"
        )

        organization.send_notice("Initializing room...")

        room = StreamRoom(
            None,
            organization.user_id,
            organization.serv,
            [organization.user_id, organization.serv.user_id],
            [],
        )
        room.name = name.lower()
        room.organization = organization
        room.organization_id = organization.id
        room.organization_name = organization.name
        room.max_backfill_amount = organization.max_backfill_amount

        result = organization.zulip.get_stream_id(name)
        room.stream_id = result.get("stream_id")

        if not room.stream_id:
            organization.send_notice(
                f"A stream with the name {name} doesn't exist or we haven't been invited to it."
            )
            return None

        room.organization = organization
        room.organization_id = organization.id
        room.organization_name = organization.name

        # stamp global member sync setting at room creation time
        room.member_sync = organization.serv.config["member_sync"]

        organization.serv.register_room(room)
        organization.rooms[room.stream_id] = room

        asyncio.ensure_future(room.create_mx(name))
        return room

    def from_config(self, config: dict) -> None:
        super().from_config(config)

        if "key" in config:
            self.key = config["key"]

        if "member_sync" in config:
            self.member_sync = config["member_sync"]

        if "stream_id" in config:
            self.stream_id = config["stream_id"]

        if self.stream_id is None:
            raise Exception("No stream_id key in config for ChannelRoom")

        # initialize lazy members dict if sync is not off
        if self.member_sync != "off":
            if self.lazy_members is None:
                self.lazy_members = {}
        else:
            self.lazy_members = None

        if "use_displaynames" in config:
            self.use_displaynames = config["use_displaynames"]

        if "allow_notice" in config:
            self.allow_notice = config["allow_notice"]

        if "topic_sync" in config:
            self.topic_sync = config["topic_sync"]

    def to_config(self) -> dict:
        return {
            **(super().to_config()),
            "key": self.key,
            "member_sync": self.member_sync,
            "stream_id": self.stream_id,
            "use_displaynames": self.use_displaynames,
            "allow_notice": self.allow_notice,
            "topic_sync": self.topic_sync,
        }

    async def create_mx(self, name):
        # handle !room names properly
        visible_name = name
        if visible_name.startswith("!"):
            visible_name = "!" + visible_name[6:]

        result = self.organization.zulip.call_endpoint(
            url=f"/streams/{self.stream_id}", method="get"
        )
        if result["result"] != "success":
            self.send_notice(f"Could not get stream by id: {result}")
            return

        restricted = None  # Invite only
        name_prefix = "🔒"

        if not result["stream"]["invite_only"]:
            # Allow space members to join
            restricted = self.organization.space.id
            name_prefix = "#"

        self.id = await self.organization.serv.create_room(
            f"{name_prefix}{visible_name} ({self.organization.name})",
            "",
            [self.organization.user_id],
            permissions=self.organization.permissions,
            restricted=restricted,
        )
        self.serv.register_room(self)
        await self.save()
        # start event queue now that we have an id
        self._queue.start()

        # attach to organization space
        if self.organization.space:
            await self.organization.space.attach(self.id)

    # topic updates from channel state replies are ignored because formatting changes
    def set_topic(self, topic: str, user_id: Optional[str] = None) -> None:
        pass

    def on_topic(self, conn, event) -> None:
        self.send_notice(f"{event.source.nick} changed the topic")
        if conn.real_nickname != event.source.nick and self.topic_sync in [
            "matrix",
            "any",
        ]:
            raise NotImplementedError(
                "Set Matrix room topic to Zulip stream description"
            )
            # super().set_topic()

    @connected
    async def _on_mx_room_topic(self, event) -> None:
        if event.sender != self.serv.user_id and self.topic_sync in ["zulip", "any"]:
            topic = re.sub(r"[\r\n]", " ", event.content.topic)
            self.organization.conn.topic(self.name, topic)

    @connected
    async def on_mx_message(self, event) -> None:
        sender = str(event.sender)
        (name, server) = sender.split(":", 1)

        # ignore self messages
        if sender == self.serv.user_id:
            return

        # prevent re-sending federated messages back
        if (
            name.startswith("@" + self.serv.puppet_prefix)
            and server == self.serv.server_name
        ):
            return

        if self.use_displaynames and event.sender in self.displaynames:
            sender_displayname = self.displaynames[event.sender][:100]

            sender_displayname = f"[{sender_displayname}](https://matrix.to/#/{sender})"

            sender = sender_displayname

        if event.content.msgtype.is_media:
            raise NotImplementedError("Sending media to Zulip")
        elif event.content.msgtype in (
            MessageType.EMOTE,
            MessageType.TEXT,
            MessageType.NOTICE,
        ):
            await self._relay_message(event, sender)

        await self.az.intent.send_receipt(event.room_id, event.event_id)

    async def _relay_message(self, event, sender):
        prefix = f"<{sender}> "

        # try to find out if this was a reply
        reply_to = None
        if event.content.get_reply_to():
            rel_event = event

            # traverse back all edits
            while rel_event.content.get_edit():
                rel_event = await self.az.intent.get_event(
                    self.id, rel_event.content.get_edit()
                )

            # see if the original is a reply
            if rel_event.content.get_reply_to():
                reply_to = await self.az.intent.get_event(
                    self.id, rel_event.content.get_reply_to()
                )

        # Get topic (Matrix thread)
        thread_id = event.content.get_thread_parent()
        topic = None

        # Reply in thread
        if thread_id is not None:
            thread_event = await self.az.intent.get_event(self.id, thread_id)
            topic = thread_event.content.body

            # Save last thread event for old clients
            self.thread_last_message[thread_event.event_id] = event.event_id

            # Save thread id for topic
            if topic not in self.threads:
                self.threads[topic] = thread_event.event_id

        if event.content.get_edit():
            message = await self._process_event_content(event, prefix, reply_to)
            self.last_messages[event.sender] = event
        else:
            # keep track of the last message
            self.last_messages[event.sender] = event
            message = await self._process_event_content(event, prefix)

        request = {
            "type": "stream",
            "to": self.stream_id,
            "topic": topic,
            "content": message,
        }
        self.organization.zulip.send_message(request)

        await self.save()

    @connected
    async def on_mx_ban(self, user_id) -> None:
        pass

    @connected
    async def on_mx_unban(self, user_id) -> None:
        pass

    @connected
    async def on_mx_leave(self, user_id) -> None:
        pass

    def pills(self):
        # if pills are disabled, don't generate any
        if self.organization.pills_length < 1:
            return None

        ret = super().pills()

        # remove the bot from pills as it may cause confusion
        nick = self.organization.conn.real_nickname.lower()
        if nick in ret:
            del ret[nick]

        return ret

    async def cmd_displaynames(self, args) -> None:
        if args.enabled is not None:
            self.use_displaynames = args.enabled
            await self.save()

        self.send_notice(
            f"Displaynames are {'enabled' if self.use_displaynames else 'disabled'}"
        )

    async def cmd_noticerelay(self, args) -> None:
        if args.enabled is not None:
            self.allow_notice = args.enabled
            await self.save()

        self.send_notice(
            f"Notice relay is {'enabled' if self.allow_notice else 'disabled'}"
        )

    async def cmd_topic(self, args) -> None:
        if args.sync is None:
            self.organization.conn.topic(self.name, " ".join(args.text))
            return

        self.topic_sync = args.sync if args.sync != "off" else None
        self.send_notice(
            f"Topic sync is {self.topic_sync if self.topic_sync else 'off'}"
        )
        await self.save()

    async def cmd_sync(self, args):
        if args.lazy:
            self.member_sync = "lazy"
            await self.save()
        elif args.half:
            self.member_sync = "half"
            await self.save()
        elif args.full:
            self.member_sync = "full"
            await self.save()
        elif args.off:
            self.member_sync = "off"
            # prevent anyone already in lazy list to be invited
            self.lazy_members = None
            await self.save()

        self.send_notice(
            f"Member sync is set to {self.member_sync}", forward=args._forward
        )

    def _add_puppet(self, zulip_user):
        mx_user_id = self.serv.get_mxid_from_zulip_user_id(
            self.organization, zulip_user["user_id"]
        )

        self.ensure_zulip_user_id(self.organization, zulip_user=zulip_user)
        self.join(mx_user_id, zulip_user["full_name"])

    def _remove_puppet(self, user_id, reason=None):
        if user_id == self.serv.user_id or user_id == self.user_id:
            return

        self.leave(user_id, reason)

    def on_join(self, zulip_user_id: int = None, zulip_user: dict = None) -> None:
        if zulip_user_id is None:
            zulip_user_id = zulip_user["user_id"]
        # we don't need to sync ourself
        if zulip_user_id == self.organization.profile["user_id"]:
            return
        if zulip_user is None:
            zulip_user = self.organization.get_zulip_user(zulip_user_id)

        # ensure, append, invite and join
        self._add_puppet(zulip_user)
        mx_user_id = self.serv.get_mxid_from_zulip_user_id(
            self.organization, zulip_user_id
        )
        self.join(mx_user_id, zulip_user["full_name"], lazy=False)

    def on_part(self, zulip_user_id) -> None:
        # we don't need to sync ourself
        if zulip_user_id == self.organization.profile["user_id"]:
            return

        mx_user_id = self.serv.get_mxid_from_zulip_user_id(
            self.organization, zulip_user_id
        )
        self._remove_puppet(mx_user_id)

    async def sync_zulip_members(self, subscribers: list):
        to_remove = []
        to_add = []

        # always reset lazy list because it can be toggled on-the-fly
        self.lazy_members = {} if self.member_sync != "off" else None

        # build to_remove list from our own puppets
        for member in self.members:
            (name, server) = member.split(":", 1)

            if (
                name.startswith("@" + self.serv.puppet_prefix)
                and server == self.serv.server_name
            ):
                to_remove.append(member)

        for zulip_user_id in subscribers:
            # convert to mx id, check if we already have them
            mx_user_id = self.serv.get_mxid_from_zulip_user_id(
                self.organization, zulip_user_id
            )

            # make sure this user is not removed from room
            if mx_user_id in to_remove:
                to_remove.remove(mx_user_id)
                continue

            # ignore adding us here, only lazy join on echo allowed
            if zulip_user_id == self.organization.profile["user_id"]:
                continue

            # if this user is not in room, add to invite list
            if not self.in_room(mx_user_id):
                to_add.append((mx_user_id, zulip_user_id))

            # always put everyone in the room to lazy list if we have any member sync
            if self.lazy_members is not None:
                self.lazy_members[mx_user_id] = zulip_user_id

        # never remove us or appservice
        if self.serv.user_id in to_remove:
            to_remove.remove(self.serv.user_id)
        if self.user_id in to_remove:
            to_remove.remove(self.user_id)

        for mx_user_id, zulip_user_id in to_add:
            zulip_user = self.organization.get_zulip_user(zulip_user_id)

            self._add_puppet(zulip_user)

        for mx_user_id in to_remove:
            self._remove_puppet(mx_user_id, "Unsubcribed from stream")

    async def backfill_messages(self):
        request = {
            "anchor": "newest",
            "num_before": self.max_backfill_amount,
            "num_after": 0,
            "narrow": [
                {"operator": "stream", "operand": self.stream_id},
            ],
        }
        result = self.organization.zulip.get_messages(request)

        if result["result"] != "success":
            logging.error(f"Failed getting Zulip messages: {result['msg']}")
            return

        for message in result["messages"]:
            if str(message["id"]) in self.organization.messages:
                continue
            self.organization.zulip_handler.backfill_message(message)



# MatrixZulipBridge - an appservice puppeting bridge for Matrix - Zulip
#
# Copyright (C) 2024 Emma Meijere <emgh@em.id.lv>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Originally licensed under the MIT (Expat) license:
# <https://github.com/hifi/heisenbridge/blob/2532905f13835762870de55ba8a404fad6d62d81/LICENSE>.
#
# [This file includes modifications made by Emma Meijere]
#
#
from matrixzulipbridge import __version__
from matrixzulipbridge.command_parse import CommandManager, CommandParserError
from matrixzulipbridge.room import Room


class PersonalRoom(Room):
    commands: CommandManager

    def init(self):
        self.commands = CommandManager()

        self.mx_register("m.room.message", self.on_mx_message)

    def is_valid(self) -> bool:
        if self.user_id is None:
            return False

        if len(self.members) != 2:
            return False

        return True

    async def show_help(self):
        self.send_notice_html(
            f"<b>Howdy, stranger!</b> This is your personal <b>{self.serv.server_name}</b> room."
        )

        try:
            return await self.commands.trigger("HELP")
        except CommandParserError as e:
            return self.send_notice(str(e))

    async def on_mx_message(self, event) -> bool:
        if str(event.content.msgtype) != "m.text" or event.sender == self.serv.user_id:
            return

        # ignore edits
        if event.content.get_edit():
            return

        try:
            lines = event.content.body.split("\n")

            command = lines.pop(0)
            tail = "\n".join(lines) if len(lines) > 0 else None

            await self.commands.trigger(command, tail)
        except CommandParserError as e:
            self.send_notice(str(e))

    async def cmd_version(self, _args):
        self.send_notice(f"zulipbridge v{__version__}")



# MatrixZulipBridge - an appservice puppeting bridge for Matrix - Zulip
#
# Copyright (C) 2024 Emma Meijere <emgh@em.id.lv>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Originally licensed under the MIT (Expat) license:
# <https://github.com/hifi/heisenbridge/blob/2532905f13835762870de55ba8a404fad6d62d81/LICENSE>.
#
# [This file includes modifications made by Emma Meijere]
#
#
import os
import shutil
import subprocess

module_dir = os.path.dirname(__file__)
root_dir = module_dir + "/../"

__version__ = "0.0.0"
__git_version__ = None

if os.path.exists(module_dir + "/version.txt"):
    __version__ = open(module_dir + "/version.txt", encoding="utf-8").read().strip()

if os.path.exists(root_dir + ".git") and shutil.which("git"):
    try:
        git_env = {
            "PATH": os.environ["PATH"],
            "HOME": os.environ["HOME"],
            "LANG": "C",
            "LC_ALL": "C",
        }
        git_bits = (
            subprocess.check_output(
                ["git", "describe", "--tags"],
                stderr=subprocess.DEVNULL,
                cwd=root_dir,
                env=git_env,
            )
            .strip()
            .decode("ascii")
            .split("-")
        )

        __git_version__ = git_bits[0][1:]

        if len(git_bits) > 1:
            __git_version__ += f".dev{git_bits[1]}"

        if len(git_bits) > 2:
            __git_version__ += f"+{git_bits[2]}"

        # always override version with git version if we have a valid version number
        __version__ = __git_version__
    except (subprocess.SubprocessError, OSError):
        pass



# MatrixZulipBridge - an appservice puppeting bridge for Matrix - Zulip
#
# Copyright (C) 2024 Emma Meijere <emgh@em.id.lv>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Originally licensed under the MIT (Expat) license:
# <https://github.com/hifi/heisenbridge/blob/2532905f13835762870de55ba8a404fad6d62d81/LICENSE>.
#
# [This file includes modifications made by Emma Meijere]
#
#
"""Pupetting Matrix - Zulip bridge"""
from matrixzulipbridge.version import __version__



# MatrixZulipBridge - an appservice puppeting bridge for Matrix - Zulip
#
# Copyright (C) 2024 Emma Meijere <emgh@em.id.lv>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Originally licensed under the MIT (Expat) license:
# <https://github.com/hifi/heisenbridge/blob/2532905f13835762870de55ba8a404fad6d62d81/LICENSE>.
#
# [This file includes modifications made by Emma Meijere]
#
#
import logging
import re
from abc import ABC
from collections import defaultdict
from typing import Callable, Optional

from mautrix.appservice import AppService as MauService
from mautrix.errors.base import IntentError
from mautrix.types import Membership
from mautrix.types.event.type import EventType

from matrixzulipbridge.appservice import AppService
from matrixzulipbridge.event_queue import EventQueue


class RoomInvalidError(Exception):
    pass


class Room(ABC):
    az: MauService
    id: str
    user_id: str
    serv: AppService
    members: list[str]
    lazy_members: Optional[dict[str, str]]
    bans: list[str]
    displaynames: dict[str, str]
    thread_last_message: dict[str, str]
    threads: dict[str, str]
    send_read_receipt: bool

    _mx_handlers: dict[str, list[Callable[[dict], bool]]]
    _queue: EventQueue

    def __init__(
        self,
        id: str,
        user_id: str,
        serv: AppService,
        members: list[str],
        bans: list[str],
    ):
        self.id = id
        self.user_id = user_id
        self.serv = serv
        self.members = list(members)
        self.bans = list(bans) if bans else []
        self.lazy_members = None
        self.displaynames = {}
        self.last_messages = defaultdict(str)
        self.thread_last_message = {}
        self.threads = {}
        self.send_read_receipt = True

        self._mx_handlers = {}
        self._queue = EventQueue(self._flush_events)

        # start event queue
        if self.id:
            self._queue.start()

        # we track room members
        self.mx_register("m.room.member", self._on_mx_room_member)

        self.init()

    @classmethod
    def init_class(cls, az: MauService):
        cls.az = az

    async def post_init(self):
        pass

    def from_config(self, config: dict) -> None:
        if "threads" in config:
            self.threads = config["threads"]

        if "send_read_receipt" in config:
            self.send_read_receipt = config["send_read_receipt"]

    def init(self) -> None:
        pass

    def is_valid(self) -> bool:
        return True

    def cleanup(self):
        self._queue.stop()

    def to_config(self) -> dict:
        return {"threads": self.threads, "send_read_receipt": self.send_read_receipt}

    async def save(self) -> None:
        config = self.to_config()
        config["type"] = type(self).__name__
        config["user_id"] = self.user_id
        await self.az.intent.set_account_data("zulip", config, self.id)

    def mx_register(self, type: str, func: Callable[[dict], bool]) -> None:
        if type not in self._mx_handlers:
            self._mx_handlers[type] = []

        self._mx_handlers[type].append(func)

    async def on_mx_event(self, event: dict) -> None:
        handlers = self._mx_handlers.get(str(event.type), [self._on_mx_unhandled_event])

        for handler in handlers:
            await handler(event)

    def in_room(self, user_id):
        return user_id in self.members

    async def on_mx_ban(self, user_id) -> None:
        pass

    async def on_mx_unban(self, user_id) -> None:
        pass

    async def on_mx_leave(self, user_id) -> None:
        pass

    async def _on_mx_unhandled_event(self, event: dict) -> None:
        pass

    async def _on_mx_room_member(self, event: dict) -> None:
        if (
            event.content.membership in [Membership.LEAVE, Membership.BAN]
            and event.state_key in self.members
        ):
            self.members.remove(event.state_key)
            if event.state_key in self.displaynames:
                del self.displaynames[event.state_key]
            if event.state_key in self.last_messages:
                del self.last_messages[event.state_key]

            if not self.is_valid():
                raise RoomInvalidError(
                    f"Room {self.id} ended up invalid after membership change, returning false from event handler."
                )

        if event.content.membership == Membership.LEAVE:
            if event.state_key in self.bans:
                self.bans.remove(event.state_key)
                await self.on_mx_unban(event.state_key)
            else:
                await self.on_mx_leave(event.state_key)

        if event.content.membership == Membership.BAN:
            if event.state_key not in self.bans:
                self.bans.append(event.state_key)

            await self.on_mx_ban(event.state_key)

        if event.content.membership == Membership.JOIN:
            if event.state_key not in self.members:
                self.members.append(event.state_key)

            if event.content.displayname is not None:
                self.displaynames[event.state_key] = str(event.content.displayname)
            elif event.state_key in self.displaynames:
                del self.displaynames[event.state_key]

    async def _join(self, user_id, nick=None):
        await self.az.intent.user(user_id).ensure_joined(self.id, ignore_cache=True)

        self.members.append(user_id)
        if nick is not None:
            self.displaynames[user_id] = nick

    async def _flush_events(self, events):
        for event in events:
            try:
                if event["type"] == "_join":
                    if event["user_id"] not in self.members:
                        await self._join(event["user_id"], event["nick"])
                elif event["type"] == "_leave":
                    if (
                        self.lazy_members is not None
                        and event["user_id"] in self.lazy_members
                    ):
                        del self.lazy_members[event["user_id"]]

                    if event["user_id"] in self.members:
                        if event["reason"] is not None:
                            await self.az.intent.user(event["user_id"]).kick_user(
                                self.id, event["user_id"], event["reason"]
                            )
                        else:
                            await self.az.intent.user(event["user_id"]).leave_room(
                                self.id
                            )
                        if event["user_id"] in self.members:
                            self.members.remove(event["user_id"])
                        if event["user_id"] in self.displaynames:
                            del self.displaynames[event["user_id"]]

                elif event["type"] == "_kick":
                    if event["user_id"] in self.members:
                        await self.az.intent.kick_user(
                            self.id, event["user_id"], event["reason"]
                        )
                        self.members.remove(event["user_id"])
                        if event["user_id"] in self.displaynames:
                            del self.displaynames[event["user_id"]]
                elif event["type"] == "_ensure_zulip_user_id":
                    await self.serv.ensure_zulip_user_id(
                        event["organization"],
                        zulip_user_id=event["zulip_user_id"],
                        zulip_user=event["zulip_user"],
                    )
                elif event["type"] == "_redact":
                    await self.az.intent.redact(
                        room_id=self.id,
                        event_id=event["event_id"],
                        reason=event["reason"],
                    )
                elif event["type"] == "_permission":
                    if len(event["content"]["users"]) == 0:
                        return  # No need to send an empty event
                    try:
                        await self.az.intent.set_power_levels(
                            room_id=self.id,
                            content=event["content"],
                        )
                    except IntentError:
                        pass
                elif "state_key" in event:
                    intent = self.az.intent

                    if event["user_id"]:
                        intent = intent.user(event["user_id"])

                    await intent.send_state_event(
                        self.id,
                        EventType.find(event["type"]),
                        state_key=event["state_key"],
                        content=event["content"],
                    )
                else:
                    bridge_data = event["content"].get("lv.shema.zulipbridge")
                    if bridge_data is None:
                        bridge_data = {}

                    if bridge_data.get("type") == "message":
                        thread_id = self.threads[bridge_data["zulip_topic"]]
                        event["content"]["m.relates_to"] = {
                            "event_id": thread_id,
                            "rel_type": "m.thread",
                        }
                        # https://spec.matrix.org/v1.9/client-server-api/#fallback-for-unthreaded-clients
                        if thread_id in self.thread_last_message:
                            event["content"]["m.relates_to"]["is_falling_back"] = True
                            event["content"]["m.relates_to"]["m.in_reply_to"] = {
                                "event_id": self.thread_last_message[thread_id]
                            }

                    intent = (
                        self.az.intent.user(event["user_id"])
                        if event["user_id"]
                        else self.az.intent
                    )

                    if "zulip_user_id" in bridge_data and "display_name" in bridge_data:
                        # TODO: Check if the display name is already cached
                        await intent.set_displayname(bridge_data["display_name"])

                    # Remove bridge data before sending it to Matrix
                    # This saves a few bytes!
                    event["content"].pop("lv.shema.zulipbridge", None)

                    timestamp = None
                    if "timestamp" in bridge_data:
                        timestamp = bridge_data["timestamp"] * 1000

                    event_type = EventType.find(event["type"])
                    event_id = await intent.send_message_event(
                        self.id,
                        event_type,
                        event["content"],
                        timestamp=timestamp,
                    )
                    if (
                        "m.relates_to" in event["content"]
                        and event["content"]["m.relates_to"]["rel_type"] == "m.thread"
                    ):
                        self.thread_last_message[
                            event["content"]["m.relates_to"]["event_id"]
                        ] = event_id

                    match bridge_data.get("type"):
                        case "message":
                            # Is this efficient?
                            self.organization.messages[
                                str(bridge_data["zulip_message_id"])
                            ] = event_id
                            await self.organization.save()

                            if self.send_read_receipt:
                                # Send read receipt to Zulip
                                self.organization.zulip.update_message_flags(
                                    {
                                        "messages": [bridge_data["zulip_message_id"]],
                                        "op": "add",
                                        "flag": "read",
                                    }
                                )

                        case "topic":
                            self.threads[bridge_data["zulip_topic"]] = event_id
                            await self.save()

            except Exception:
                logging.exception("Queued event failed")

    # send message to mx user (may be puppeted)
    def send_message(
        self,
        text: str,
        user_id: Optional[str] = None,
        formatted=None,
        fallback_html: Optional[str] = None,
        thread_id: Optional[str] = None,
        custom_data: Optional[dict] = None,
    ) -> None:
        if formatted:
            event = {
                "type": "m.room.message",
                "content": {
                    "msgtype": "m.text",
                    "format": "org.matrix.custom.html",
                    "body": text,
                    "formatted_body": formatted,
                },
                "user_id": user_id,
                "fallback_html": fallback_html,
            }
        else:
            event = {
                "type": "m.room.message",
                "content": {
                    "msgtype": "m.text",
                    "body": text,
                },
                "user_id": user_id,
                "fallback_html": fallback_html,
            }

        if thread_id:
            event["content"]["m.relates_to"] = {
                "event_id": thread_id,
                "rel_type": "m.thread",
            }

        if custom_data is not None:
            event["content"]["lv.shema.zulipbridge"] = custom_data

        if "lv.shema.zulipbridge" in event["content"]:
            bridge_data: dict = event["content"]["lv.shema.zulipbridge"]
            if bridge_data["type"] == "message":
                self._ensure_thread_for_topic(bridge_data.copy(), user_id)

        self._queue.enqueue(event)

    def redact(self, event_id: str, reason: Optional[str] = None) -> None:
        event = {"type": "_redact", "event_id": event_id, "reason": reason}

        self._queue.enqueue(event)

    def _ensure_thread_for_topic(
        self, bridge_data: dict, mx_user_id: Optional[str] = None
    ) -> Optional[str]:
        zulip_topic = bridge_data["zulip_topic"]

        if zulip_topic in self.threads:
            return self.threads[zulip_topic]

        bridge_data["type"] = "topic"

        ## Create a new thread
        logging.debug(f"Creating new thread for topic: {zulip_topic}")
        # Send topic name as a message
        event = {
            "type": "m.room.message",
            "content": {
                "msgtype": "m.text",
                "body": zulip_topic,
                "lv.shema.zulipbridge": bridge_data,
            },
            "user_id": mx_user_id,
        }
        self._queue.enqueue(event)
        self.threads[zulip_topic] = None
        return None

    # send emote to mx user (may be puppeted)
    def send_emote(
        self,
        text: str,
        user_id: Optional[str] = None,
        fallback_html: Optional[str] = None,
    ) -> None:
        event = {
            "type": "m.room.message",
            "content": {
                "msgtype": "m.emote",
                "body": text,
            },
            "user_id": user_id,
            "fallback_html": fallback_html,
        }

        self._queue.enqueue(event)

    # send notice to mx user (may be puppeted)
    def send_notice(
        self,
        text: str,
        user_id: Optional[str] = None,
        formatted=None,
        fallback_html: Optional[str] = None,
    ) -> None:
        if formatted:
            event = {
                "type": "m.room.message",
                "content": {
                    "msgtype": "m.notice",
                    "format": "org.matrix.custom.html",
                    "body": text,
                    "formatted_body": formatted,
                },
                "user_id": user_id,
                "fallback_html": fallback_html,
            }
        else:
            event = {
                "type": "m.room.message",
                "content": {
                    "msgtype": "m.notice",
                    "body": text,
                },
                "user_id": user_id,
                "fallback_html": fallback_html,
            }

        self._queue.enqueue(event)

    # send notice to mx user (may be puppeted)
    def send_notice_html(self, text: str, user_id: Optional[str] = None) -> None:
        event = {
            "type": "m.room.message",
            "content": {
                "msgtype": "m.notice",
                "body": re.sub("<[^<]+?>", "", text),
                "format": "org.matrix.custom.html",
                "formatted_body": text,
            },
            "user_id": user_id,
        }

        self._queue.enqueue(event)

    def react(self, event_id: str, text: str, user_id: Optional[str] = None) -> None:
        event = {
            "type": "m.reaction",
            "content": {
                "m.relates_to": {
                    "event_id": event_id,
                    "key": text,
                    "rel_type": "m.annotation",
                }
            },
            "user_id": user_id,
        }

        self._queue.enqueue(event)

    def set_topic(self, topic: str, user_id: Optional[str] = None) -> None:
        event = {
            "type": "m.room.topic",
            "content": {
                "topic": topic,
            },
            "state_key": "",
            "user_id": user_id,
        }

        self._queue.enqueue(event)

    def join(self, user_id: str, nick=None, lazy=False) -> None:
        event = {
            "type": "_join",
            "content": {},
            "user_id": user_id,
            "nick": nick,
            "lazy": lazy,
        }

        self._queue.enqueue(event)

    def leave(self, user_id: str, reason: Optional[str] = None) -> None:
        event = {
            "type": "_leave",
            "content": {},
            "reason": reason,
            "user_id": user_id,
        }

        self._queue.enqueue(event)

    def rename(self, old_nick: str, new_nick: str) -> None:
        event = {
            "type": "_rename",
            "content": {},
            "old_nick": old_nick,
            "new_nick": new_nick,
        }

        self._queue.enqueue(event)

    def kick(self, user_id: str, reason: str) -> None:
        event = {
            "type": "_kick",
            "content": {},
            "reason": reason,
            "user_id": user_id,
        }

        self._queue.enqueue(event)

    def ensure_zulip_user_id(self, organization, zulip_user_id=None, zulip_user=None):
        event = {
            "type": "_ensure_zulip_user_id",
            "content": {},
            "organization": organization,
            "zulip_user": zulip_user,
            "zulip_user_id": zulip_user_id,
        }

        self._queue.enqueue(event)

    async def sync_permissions(self, permissions):
        room_power_levels = await self.az.intent.get_power_levels(
            self.id, ensure_joined=False
        )

        permissions = room_power_levels.users | permissions

        if permissions == room_power_levels.users:
            logging.debug(f"Nothing chnaged: {permissions=}")
            return  # Nothing changed

        self._queue.enqueue(
            {
                "type": "_permission",
                "content": {
                    "users": permissions,
                },
            }
        )



# MatrixZulipBridge - an appservice puppeting bridge for Matrix - Zulip
#
# Copyright (C) 2024 Emma Meijere <emgh@em.id.lv>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Originally licensed under the MIT (Expat) license:
# <https://github.com/hifi/heisenbridge/blob/2532905f13835762870de55ba8a404fad6d62d81/LICENSE>.
#
# [This file includes modifications made by Emma Meijere]
#
#
import logging
import re
from typing import TYPE_CHECKING, Optional

import emoji
from markdownify import markdownify

if TYPE_CHECKING:
    from matrixzulipbridge.organization_room import OrganizationRoom
    from matrixzulipbridge.stream_room import StreamRoom


class ZulipEventHandler:
    def __init__(self, organization: "OrganizationRoom") -> None:
        self.organization = organization

    def on_event(self, event: dict):
        match event["type"]:
            case "message":
                self._handle_message(event["message"])
            case "subscription":
                self._handle_subscription(event)
            case "reaction":
                self._handle_reaction(event)
            case "delete_message":
                self._handle_delete_message(event)
            case "realm_user":
                self._handle_realm_user(event)
            case _:
                logging.debug(f"Unhandled event type: {event['type']}")

    def backfill_message(self, message: dict):
        self._handle_message(message)

    def _handle_message(self, event: dict):
        zulip_user_id = event["sender_id"]
        if zulip_user_id == self.organization.profile["user_id"]:
            return  # Ignore own messages
        if "stream_id" not in event:
            return  # Ignore DMs

        room = self._get_room_by_stream_id(event["stream_id"])

        if not room:
            logging.debug(
                f"Received message from stream with no associated Matrix room: {event}"
            )
            return

        mx_user_id = room.serv.get_mxid_from_zulip_user_id(
            self.organization, zulip_user_id
        )

        formatted_message: str = event["content"]
        formatted_message = emoji.emojize(formatted_message, language="alias")

        message = markdownify(formatted_message).rstrip()

        topic = event["subject"]

        custom_data = {
            "zulip_topic": topic,
            "zulip_user_id": zulip_user_id,
            "display_name": event["sender_full_name"],
            "zulip_message_id": event["id"],
            "type": "message",
            "timestamp": event["timestamp"],
        }

        room.send_message(
            message,
            formatted=formatted_message,
            user_id=mx_user_id,
            custom_data=custom_data,
        )

    def _handle_reaction(self, event: dict):
        zulip_user_id = event["user_id"]

        message_mxid = self.organization.messages.get(event["message_id"])
        if message_mxid is None:
            logging.warning(
                f"Could not find message with Zulip ID: {event['message_id']}, it probably wasn't sent to Matrix."
            )
            return
        # TODO: Implement

    def _handle_delete_message(self, event: dict):
        message_mxid = self._get_mxid_from_zulip_id(event["message_id"])
        if not message_mxid:
            return
        room = self._get_room_by_stream_id(event["stream_id"])
        room.redact(message_mxid, reason="Deleted on Zulip")
        del self.organization.messages[str(event["message_id"])]

    def _handle_subscription(self, event: dict):
        if not "stream_ids" in event:
            return
        for stream_id in event["stream_ids"]:
            room = self._get_room_by_stream_id(stream_id)

            if not room:
                logging.debug(
                    f"Received message from stream with no associated Matrix room: {event}"
                )
                return

            match event["op"]:
                case "peer_add":
                    for user_id in event["user_ids"]:
                        room.on_join(user_id)
                case "peer_remove":
                    for user_id in event["user_ids"]:
                        room.on_part(user_id)

    def _handle_realm_user(self, event: dict):
        # Update Zulip user cache
        if event["op"] == "update":
            user_id = event["person"]["user_id"]
            if not user_id in self.organization.zulip_users:
                return
            self.organization.zulip_users[user_id] |= event["person"]

    def _get_mxid_from_zulip_id(self, zulip_id: int | str):
        try:
            return self.organization.messages[str(zulip_id)]
        except KeyError:
            logging.debug(
                f"Message with Zulip ID {zulip_id} not found, it probably wasn't sent to Matrix"
            )

    def _get_room_by_stream_id(self, stream_id: int) -> Optional["StreamRoom"]:
        for room in self.organization.rooms.values():
            if room.stream_id == stream_id:
                return room
        return None
